package CelBot::Channel;

use strict;

use CelBot::RateLimiter;

sub new
{
   my $class = shift;
   my ( $core, $name, $config ) = @_;

   my $irc = $core->get_plugin( "irc" );

   my $self = bless {
      core => $core,
      irc  => $irc,

      name => $name,

      modes => {},

      occupants => {}, # {$nick} -> [ $flag, $jointime ]

      topic => undef,

      deferred_modes => [],
   }, $class;

   $self->reconfigure( $config );

   return $self;
}

sub reconfigure
{
   my $self = shift;
   my ( $config ) = @_;

   my $rate_count = defined $config ? $config->get_string( '@rate_count', default => 10 ) : 10;
   my $rate_time  = defined $config ? $config->get_string( '@rate_time',  default => 10 ) : 10;

   $self->{key} = defined $config ? $config->get_string( '@key', default => undef ) : undef;

   if( !$self->{ratelimiter} ) {
      my $core = $self->{core};
      my $loop = $core->get_loop;

      $self->{ratelimiter} = CelBot::RateLimiter->new( $loop, $rate_count, $rate_time );
   }
   else {
      $self->{ratelimiter}->{count} = $rate_count;
      $self->{ratelimiter}->{delay} = $rate_time;
   }
}

use overload '""' => "STRING";

sub STRING
{
   my $self = shift;
   return __PACKAGE__."[name=$self->{name}]";
}

sub name
{
   my $self = shift;
   return $self->{name};
}

sub is_joined
{
   my $self = shift;
   return $self->{is_joined};
}

sub get_modes
{
   my $self = shift;
   return $self->{modes};
}

sub get_userflags
{
   my $self = shift;

   my $occupants = $self->{occupants};

   return { map { $_ => $occupants->{$_}[0] } keys %$occupants };
}

sub get_userflag
{
   my $self = shift;
   my ( $user ) = @_;

   my $occupants = $self->{occupants};

   my $nick_folded = $user->nick_folded;

   if( exists $occupants->{$nick_folded} ) {
      return $occupants->{$nick_folded}[0];
   }
   else {
      return undef;
   }
}

sub get_jointime
{
   my $self = shift;
   my ( $user ) = @_;

   my $occupants = $self->{occupants};

   my $nick_folded = $user->nick_folded;

   if( exists $occupants->{$nick_folded} ) {
      return $occupants->{$nick_folded}[1];
   }
   else {
      return undef;
   }
}

sub get_users
{
   my $self = shift;

   my $occupants = $self->{occupants};

   my $irc = $self->{irc};
   my @users = map { $irc->get_user( $_ ) } keys %$occupants;

   return @users;
}

sub join
{
   my $self = shift;
   my %args = @_;

   my $name = $self->{name};
   my $irc = $self->{irc};

   $irc->send_message( "JOIN", undef, $name, defined $self->{key} ? ( $self->{key} ) : () );

   $self->{on_joined} = $args{on_joined};
}

sub kick
{
   my $self = shift;
   my ( $user, $message ) = @_;

   my $name = $self->{name};
   my $irc = $self->{irc};

   $irc->send_message( "KICK", undef, $self->name, $user->nick, $message );
}

sub mode
{
   my $self = shift;
   my ( $flag, $arg, $defer ) = @_;

   if( $defer ) {
      push @{ $self->{deferred_modes} }, [ $flag, $arg ];
      if( !$self->{flush_timer_id} ) {
         my $core = $self->{core};
         my $loop = $core->get_loop;
         $self->{flush_timer_id} = $loop->enqueue_timer(
            delay => 2,
            code => sub {
               undef $self->{flush_timer_id};
               $self->_flush_modes;
            },
         );
      }
   }
   else {
      my $irc = $self->{irc};
      $irc->send_message( "MODE", undef, $self->name, $flag, ( defined $arg ? ( $arg ) : () ) );
   }
}

sub _flush_modes
{
   my $self = shift;

   my @modes = splice @{ $self->{deferred_modes} };
   return unless @modes;

   my $irc = $self->{irc};

   my $flags = "";
   my @args;

   my $maxmodes = $irc->isupport( "MODES" ) || 3;

   my $lastsense = "";
   my $count = 0;
   foreach my $m ( @modes ) {
      my ( $flag, $arg ) = @$m;
      ( my $sense, $flag ) = $flag =~ m/^(.)(.*)/;

      if( $sense ne $lastsense ) {
         $flags .= $sense;
         $lastsense = $sense;
      }

      $flags .= $flag;
      push @args, $arg if defined $arg;

      $count++;

      if( $count == $maxmodes ) {
         $irc->send_message( "MODE", undef, $self->name, $flags, @args );
         $flags = "";
         @args = ();
         $lastsense = "";
         $count = 0;
      }
   }

   if( $flags ) {
      $irc->send_message( "MODE", undef, $self->name, $flags, @args );
   }
}

sub part
{
   my $self = shift;
   my %args = @_;

   my $name = $self->{name};
   my $irc = $self->{irc};

   $irc->send_message( "PART", undef, $name );

   $self->{on_parted} = $args{on_parted};
}

sub privmsg
{
   my $self = shift;
   my ( $text ) = @_;

   my $irc = $self->{irc};

   $self->{ratelimiter}->do( sub {
      $irc->send_message( "PRIVMSG", undef, $self->{name}, $text );
   } );
}

sub notice
{
   my $self = shift;
   my ( $text ) = @_;

   my $irc = $self->{irc};

   $self->{ratelimiter}->do( sub {
      $irc->send_message( "NOTICE", undef, $self->{name}, $text );
   } );
}

sub restricted_privmsg
{
   my $self = shift;
   my ( $text, $restriction ) = @_;

   my $irc = $self->{irc};

   $self->{ratelimiter}->do( sub {
      $irc->send_message( "PRIVMSG", undef, $restriction . $self->{name}, $text );
   } );
}

sub start_who
{
   my $self = shift;

   my $irc = $self->{irc};
   $irc->send_message( "WHO", undef, $self->{name} );
}

# There are three ways a user could leave a channel, PART, QUIT, KICK. Handle
# them all here
sub user_leave
{
   my $self = shift;
   my ( $user, $command, $reason, $kicker ) = @_;

   my $occupants = $self->{occupants};

   delete $occupants->{$user->nick_folded};

   my $core = $self->{core};
   $core->run_plugins( "on_channel_leave", $self, $user, $command, $reason, $kicker );
}

sub user_rename
{
   my $self = shift;
   my ( $oldname, $newname ) = @_;

   my $occupants = $self->{occupants};

   return unless exists $occupants->{$oldname};

   my $occ = delete $occupants->{$oldname};
   $occupants->{$newname} = $occ;

   my $core = $self->{core};
   $core->run_plugins( "on_channel_user_rename", $self, $oldname, $newname );
}

sub on_message_311 # No TOPIC
{
   my $self = shift;
   my ( $command, $message ) = @_;

   $self->{topic} = undef;

   my $core = $self->{core};
   $core->run_plugins( "on_channel_topic", $self, $self->{topic} );

   return 1;
}

sub on_message_332 # TOPIC
{
   my $self = shift;
   my ( $command, $message ) = @_;

   $self->{topic} = $message->arg(2);

   my $core = $self->{core};
   $core->run_plugins( "on_channel_topic", $self, $self->{topic} );

   return 1;
}

sub on_message_names
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   my $names = $hints->{names};

   my $irc = $self->{irc};

   my %occupants;

   foreach my $nick_folded ( keys %$names ) {
      my $flag = $names->{$nick_folded}{flag};

      # Make sure the User object exists
      $irc->nick_to_user( $names->{$nick_folded}{nick} );

      $occupants{$nick_folded} = [ $flag, 0 ]; # put 0 since we don't know the jointime
   }

   $self->{occupants} = \%occupants;

   my $core = $self->{core};
   $core->run_plugins( "on_channel_names", $self, $self->get_userflags );

   return 1;
}

sub on_message_who
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   my $irc = $self->{irc};

   my @who;
   foreach ( @{ $hints->{who} } ) {
      my ( $nick, $ident, $host, $flags ) = @{$_}{qw( user_nick user_ident user_host user_flags )};

      my $user = $irc->nick_to_user( $nick );
      $user->advise_host( $ident, $host );

      push @who, {
         nick   => $nick,
         ident  => $ident,
         host   => $host,
         flags  => $flags,
      };
   }

   my $core = $self->{core};
   $core->run_plugins( "on_channel_who", $self, @who );

   return 1;
}

sub on_message_JOIN
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   if( $hints->{prefix_is_me} ) {
      $self->{is_joined} = 1;

      $self->{on_joined}->() if $self->{on_joined};
      undef $self->{on_joined};

      $self->{names} = [];

      my $core = $self->{core};
      $core->run_plugins( "on_channel_self_join", $self );

      return 1;
   }

   my $prefix = $message->prefix;

   my $irc = $self->{irc};
   my $user = $irc->hints_to_user( $hints );

   my $occupants = $self->{occupants};

   $occupants->{$user->nick_folded} = [ "", time() ];

   my $core = $self->{core};
   $core->run_plugins( "on_channel_join", $self, $user );

   return 1;
}

sub on_message_KICK
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   my $kicked = $message->arg(1);

   my $reason = $message->arg(2);
   defined $reason or $reason = "";

   my $irc = $self->{irc};
   my $kickeduser = $irc->nick_to_user( $kicked );

   my $core = $self->{core};

   if( $kickeduser->is_me ) {
      $self->{is_joined} = 0;

      $core->run_plugins( "on_channel_self_leave", $self, "KICK", $reason );

      return 1;
   }

   $self->user_leave( $kickeduser, "KICK", $reason, $irc->nick_to_user( $hints->{kicker_nick} ) );

   return 1;
}

sub on_message_MODE
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   my $core = $self->{core};
   my $irc = $self->{irc};

   my $modes = $self->{modes};

   my ( $hostmodes, $opaquemodes, $nummodes, $boolmodes ) = @{ $irc->isupport( "chanmodes_list" ) };

   my ( undef, $modeflags, @modeargs ) = $message->args;

   my $occupants = $self->{occupants};

   my $sense = 0;
   foreach my $modechar ( split( m//, $modeflags ) ) {
      $sense =  1, next if $modechar eq "+";
      $sense = -1, next if $modechar eq "-";

      if( my $flag = $irc->prefix_mode2flag( $modechar ) ) {
         my $nick = $irc->casefold_name( shift @modeargs );

         if( $sense > 0 ) {
            $occupants->{$nick}[0] = $flag;
         }
         else {
            $occupants->{$nick}[0] = "";
         }

         my $user = $irc->nick_to_user( $nick );

         my $core = $self->{core};
         $core->run_plugins( "on_channel_usermode", $self, $user, $sense>0 ? $flag : "" );
      }
      elsif( index($hostmodes, $modechar) > -1 ) {
         my $target = shift @modeargs;
         $core->log( "DEBUG", "Channel[$self->{name}]: ", "MODE HOSTMASK $target $sense $modechar" );
      }
      elsif( index($nummodes, $modechar) > -1 or index($opaquemodes, $modechar) > -1 ) {
         my $value;
         if( $sense > 0 ) {
            $value = shift @modeargs;
            $modes->{$modechar} = $value;
         }
         else {
            $value = undef;
            delete $modes->{$modechar};
         }

         my $core = $self->{core};
         $core->run_plugins( "on_channel_mode", $self, $modechar, $value );
      }
      elsif( index($boolmodes, $modechar) > -1 ) {
         my $value;
         if( $sense > 0 ) {
            $value = 1;
            $modes->{$modechar} = $value;
         }
         else {
            $value = undef;
            delete $modes->{$modechar};
         }

         my $core = $self->{core};
         $core->run_plugins( "on_channel_mode", $self, $modechar, $value );
      }
      else {
         $core->log( "DEBUG", "Channel[$self->{name}]: ", "Unrecognised MODE $modechar" );
      }
   }

   return 1;
}

sub on_message_PART
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   if( $hints->{prefix_is_me} ) {
      $self->{is_joined} = 0;

      my $core = $self->{core};
      $core->run_plugins( "on_channel_self_leave", $self, "PART" );

      $self->{on_parted}->() if $self->{on_parted};
      undef $self->{on_parted};

      return 1;
   }

   my $reason = $message->arg(1);
   defined $reason or $reason = "";

   my $irc = $self->{irc};
   my $user = $irc->hints_to_user( $hints );

   $self->user_leave( $user, "PART", $reason );

   return 1;
}

sub on_message_TOPIC
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   $self->{topic} = $message->arg(1);

   my $core = $self->{core};
   $core->run_plugins( "on_channel_topic", $self, $self->{topic}, $message->prefix );

   return 1;
}

sub on_message_text
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   my $irc = $self->{irc};

   my $user = $irc->hints_to_user( $hints );

   my $core = $self->{core};
   $core->run_plugins( "on_channel_privmsg", $self, $user, $message->arg(1) );

   return 1;
}

sub on_message_ctcp_ACTION
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   my $irc = $self->{irc};

   my $user = $irc->hints_to_user( $hints );

   my $core = $self->{core};
   $core->run_plugins( "on_channel_ctcp_ACTION", $self, $user, $hints->{ctcp_args} );

   return 1;
}

# Keep perl happy, keep Britain tidy
1;
