package CelBot::Channels;

use strict;

use CelBot::Channel;
use CelBot::Commands;

sub new
{
   my $class = shift;
   my ( $core, $config ) = @_;

   my $irc = $core->get_plugin( "irc" );

   my $self = bless {
      core => $core,
      irc  => $irc,
   }, $class;

   my %channels;

   $config->associate_nodeset(
      'channel',
      '@name',

      add => sub {
         my ( $name, $config ) = @_;

         $name = $irc->casefold_name( $name );

         # Deal with adding config to channels that were joined by command
         if( exists $channels{$name} ) {
            $channels{$name}->reconfigure( $config ) if $channels{$name}->can( "reconfigure" );
         }
         else {
            $channels{$name} = CelBot::Channel->new( $core, $name, $config );

            if( $irc->is_connected ) {
               $channels{$name}->join;
            }
         }
      },

      keep => sub {
         my ( $name, $config ) = @_;

         $name = $irc->casefold_name( $name );

         $channels{$name}->reconfigure( $config ) if $channels{$name}->can( "reconfigure" );

         $config->reload;
      },

      remove => sub {
         my ( $name ) = @_;

         $channels{$name}->part(
            on_parted => sub { delete $channels{$name} }
         );
      },
   );

   $self->{channels} = \%channels;

   return $self;
}

sub get_channel
{
   my $self = shift;
   my ( $name ) = @_;

   my $irc = $self->{irc};
   $name = $irc->casefold_name( $name );

   return exists $self->{channels}->{$name} ? $self->{channels}->{$name}
                                            : undef;
}

sub get_channels
{
   my $self = shift;
   return values %{ $self->{channels} };
}

sub gen_argspec
{
   my $self = shift;

   if( @_ ) {
      my $name = shift;
      return CelBot::Channels::ArgSpec->new( $name, channels => $self, @_ );
   }
   else {
      # Cache a single one
      return $self->{argspec} ||= CelBot::Channels::ArgSpec->new( 'channel', channels => $self );
   }
}

sub on_message
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   return 0 unless $hints->{target_type} and $hints->{target_type} eq "channel";

   if( my $chan = $self->{channels}->{ $hints->{target_name_folded} } ) {
      # Turn ctcp ACTION into ctcp_ACTION, etc
      $command =~ s/ /_/g;
      my $method = "on_message_" . $command;
      return 1 if $chan->can( $method ) and $chan->$method( $command, $message, $hints );

      return 1 if $chan->can( "on_message" ) and $chan->on_message( $command, $message, $hints );
   }

   return 0;
}

sub on_message_001
{
   my $self = shift;

   foreach my $chan ( values %{ $self->{channels} } ) {
      $chan->join;
   }

   return 1;
}

sub on_user_rename
{
   my $self = shift;
   my ( $user, $oldnick, $newnick ) = @_;

   my $irc = $self->{irc};

   my $old_folded = $irc->casefold_name( $oldnick );
   my $new_folded = $irc->casefold_name( $newnick );

   return 1 if $old_folded eq $new_folded;

   foreach my $chan ( values %{ $self->{channels} } ) {
      $chan->user_rename( $old_folded, $new_folded );
   }

   return 1;
}

sub on_user_quit
{
   my $self = shift;
   my ( $user, $reason ) = @_;

   foreach my $chan ( values %{ $self->{channels} } ) {
      defined $chan->get_userflag( $user ) or next;
      $chan->user_leave( $user, "QUIT", $reason );
   }

   return 1;
}

###
# Commands
###

sub register_commands
{
   my $self = shift;
   my ( $commands_plugin ) = @_;

   $commands_plugin->register(
      plugin  => $self,
      command => "in",
      perm    => 'public',

      args => [
         $self->gen_argspec,
         CelBot::Commands::ArgSpec->new( 'command', trailing => 1 ),
      ],

      summary => "Execute a command in the context of a (different) channel",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "join",
      perm    => 'master',

      args => [
         CelBot::Commands::ArgSpec::Bareword->new( 'channel' ),
      ],

      summary => "Join a channel",
   );

   $commands_plugin->register(
      plugin  => $self,
      scope   => 'channel|privmsg',
      command => "leave",
      perm    => 'master',

      args => [
         $self->gen_argspec( 'channel', optional => 1 ),
      ],

      summary => "Leave a channel",
   );

   $commands_plugin->register(
      plugin  => $self,
      scope   => 'channel|privmsg',
      command => "rejoin",
      perm    => 'master',

      args => [
         $self->gen_argspec( 'channel', optional => 1 ),
      ],

      summary => "Leave then rejoin a channel",
   );
}

sub command_in
{
   my $self = shift;
   my ( $context, $channel ) = @_;

   my $subcontext = $context->new_sub( 
      scope   => "channel",
      channel => $channel,
   );

   my $core = $self->{core};

   my $cmds_plugin = $core->get_plugin( "commands" );
   $cmds_plugin->run_command( $subcontext );

   return ();
}

sub command_join
{
   my $self = shift;
   my ( $context, $channame ) = @_;

   my $channel = $self->get_channel( $channame );

   if( !defined $channel ) {
      $channel = CelBot::Channel->new( $self->{core}, $channame, undef );
      $self->{channels}->{$channame} = $channel;
   }

   if( $channel->is_joined ) {
      return ( "Already joined '$channame'" );
   }

   $channel->join( 
      on_joined => sub {
         $context->respond( "Joined '$channame'" );
      }
   );

   return ();
}

sub command_leave
{
   my $self = shift;
   my ( $context, $channel ) = @_;

   $channel ||= $context->{channel};

   if( !defined $channel ) {
      return ( "Need to supply 'channel'" );
   }

   my $channame = $channel->name;

   $channel->part( 
      on_parted => sub {
         $context->respond( "Left '$channame'" );
      }
   );

   return ();
}

sub command_rejoin
{
   my $self = shift;
   my ( $context, $channel ) = @_;

   $channel ||= $context->{channel};

   if( !defined $channel ) {
      return ( "Need to supply 'channel'" );
   }

   my $channame = $channel->name;

   $channel->part( 
      on_parted => sub {
         $channel->join( 
            on_joined => sub {
               $context->respond( "Rejoined '$channame'" );
            }
         );
      }
   );

   return ();
}

# Keep perl happy, keep Britain tidy
1;


package CelBot::Channels::ArgSpec;

use strict;
use base qw( CelBot::Commands::ArgSpec::Bareword );

sub validate
{
   my $self = shift;
   my ( $channame ) = @_;

   my $channel = $self->{channels}->get_channel( $channame );

   die "No such channel '$channame'\n" if !defined $channel;

   return $channel;
}

1;
