package CelBot::Core;

use strict;

use CelBot::Plugins qw( construct_plugin );

# We need an ordered hash. We'll cheat

my @builtins = qw(
   irc
   commands
   users
   channels
   control
   settings
);

my %builtin_package = (
   irc      => 'CelBot::IRC',
   commands => 'CelBot::Commands',
   users    => 'CelBot::Users',
   channels => 'CelBot::Channels',
   control  => 'CelBot::Control',
   settings => 'CelBot::Settings',
);

sub new
{
   my $class = shift;
   my %args = @_;

   my $config = $args{config};

   my $tag = $config->get_string( '@tag' );

   my %plugins;

   my $self = bless {
      config => $config,
      loop   => $args{loop},

      tag  => $tag,

      plugins => \%plugins,

      isupport => {},
   }, $class;

   $self->reconfigure( $config );

   # Then the dynamic ones from config
   $config->associate_nodeset(
      'plugin',
      '@type',

      add => sub {
         my ( $type, $config ) = @_;

         $self->load_plugin( $type, $config );
      },

      keep => sub {
         my ( $type, $config ) = @_;

         my $plugin = $self->get_plugin( $type );

         $plugin->reconfigure( $config ) if $plugin->can( "reconfigure" );

         $config->reload;
      },

      remove => sub {
         my ( $type ) = @_;

         delete $self->{plugins}->{$type};
      },
   );

   # Reconfigure the commands module for the first time, so it collects
   # command registrations
   $plugins{commands}->reconfigure;

   $self->run_plugins( "connect" );

   return $self;
}

use overload '""' => "STRING";

sub STRING
{
   my $self = shift;
   return __PACKAGE__."[tag=$self->{tag}]";
}

sub reconfigure
{
   my $self = shift;
   my ( $config ) = @_;

   # Now reconfigure the standard builtin plugins
   for my $builtin_name ( @builtins ) {
      if( exists $self->{plugins}->{$builtin_name} ) {
         my $builtin = $self->{plugins}->{$builtin_name};
         $builtin->reconfigure( $config ) if $builtin->can( "reconfigure" );
      }
      else {
         my $builtin_class = $builtin_package{$builtin_name};
         ( my $builtin_file = "$builtin_class.pm" ) =~ s{::}{/};

         require $builtin_file;

         $self->{plugins}->{$builtin_name} = $builtin_class->new( $self, $config );
      }
   }
}

sub load_plugin
{
   my $self = shift;
   my ( $type, $config ) = @_;

   my $plugin = construct_plugin( $type, $self, $config );

   $self->{plugins}->{$type} = $plugin;
}

sub get_plugin
{
   my $self = shift;
   my ( $type ) = @_;

   return $self->{plugins}->{$type} if exists $self->{plugins}->{$type};
   return undef;
}

sub require_plugin
{
   my $self = shift;
   my ( $type ) = @_;

   return $self->{plugins}->{$type} if exists $self->{plugins}->{$type};
   return $self->load_plugin( $type, undef );
}

sub plugins
{
   my $self = shift;
   return values %{ $self->{plugins} };
}

sub run_plugins
{
   my $self = shift;
   my ( $methods, @args ) = @_;

   my $handled = 0;

   my @methods = ref $methods eq "ARRAY" ? @$methods : ( $methods );

   my @deferred;

   {
      local $self->{deferral_queue} = [];

      foreach my $plugin ( $self->plugins ) {
         METHOD: foreach my $method ( @methods ) {
            if( $plugin->can( $method ) ) {
               my $ret = eval { $plugin->$method( @args ) };

               if( !defined $ret and $@ ) {
                  $self->err( "Plugin $plugin failed - $@" );
               }

               $handled = 1 if $ret;

               last METHOD;
            }
         }
      }

      @deferred = @{ $self->{deferral_queue} };
   }

   foreach my $d ( @deferred ) {
      $self->run_plugins( @$d );
   }

   return $handled;
}

sub run_plugins_after
{
   my $self = shift;
   my ( $methods, @args ) = @_;

   if( $self->{deferral_queue} ) {
      push @{ $self->{deferral_queue} }, [ $methods, @args ];
   }
   else {
      # Just run it directly
      $self->run_plugins( $methods, @args );
   }
}

sub ask_plugins
{
   my $self = shift;
   my ( $method, @args ) = @_;

   foreach my $plugin ( $self->plugins ) {
      if( $plugin->can( $method ) ) {
         my $ret = eval { $plugin->$method( @args ) };

         if( !defined $ret and $@ ) {
            $self->err( "Plugin $plugin failed - $@" );
         }

         return $ret if defined $ret;
      }
   }

   return undef;
}

sub ask_plugins_list
{
   my $self = shift;
   my ( $method, @args ) = @_;

   my @answer;

   foreach my $plugin ( $self->plugins ) {
      if( $plugin->can( $method ) ) {
         my @ret = eval { $plugin->$method( @args ) };

         if( $@ ) {
            $self->err( "Plugin $plugin failed - $@" );
         }

         push @answer, @ret;
      }
   }

   return @answer;
}

sub scatter_plugins
{
   my $self = shift;
   my ( $on_finished, $method, @args ) = @_;

   my %results;
   my %later;
   my $ready = 0;

   foreach my $pluginname ( keys %{ $self->{plugins} } ) {
      my $plugin = $self->{plugins}->{$pluginname};
      if( $plugin->can( $method ) ) {
         my $will_async = 0;

         my $gencb = sub {
            $will_async = 1;

            $later{$pluginname} = 1;

            return sub {
               $results{$pluginname} = shift;
               delete $later{$pluginname};
               $on_finished->( %results ) if $ready and not keys %later;
            };
         };

         my $ret = eval { $plugin->$method( $gencb, @args ) };

         if( $@ ) {
            $self->err( "Plugin $plugin failed - $@" );
            delete $later{$pluginname}; # So we won't wait for it
            next;
         }

         if( not $will_async and defined $ret ) {
            $results{$pluginname} = $ret;
         }
      }
   }

   $ready = 1;
   $on_finished->( %results ) if not keys %later;
}

sub get_loop
{
   my $self = shift;
   return $self->{loop};
}

# One shared console
our $console;
our $tickit;

sub get_console
{
   my $self = shift;

   return $console if $console;

   require Tickit::Async;
   require CelBot::Console;

   $tickit = Tickit::Async->new;
   $self->get_loop->add( $tickit );

   $console = CelBot::Console->new;
   $tickit->set_root_widget( $console );

   $tickit->setup_term;

   my $globaltab = $console->{globaltab};

   $SIG{__WARN__} = sub {
      return unless defined $console;
      my $message = $_[0];
      chomp $message;
      $globaltab->add_line( "WARN: $message", indent => 6 );
   };

   return $console;
}

sub log
{
   my $self = shift;
   my ( $tag, $subject, $message ) = @_;

   chomp $message;

   $self->run_plugins( "do_log", $tag, $subject, $message );
}

sub err
{
   my $self = shift;
   my ( $message ) = @_;

   $self->run_plugins( "do_log", "ERR", "ERR", $message );
}

# Keep perl happy; keep Britain tidy
1;
