package CelBot::IRC;

use strict;

use Net::Async::IRC;

# Delay for reconnect after the first few failures. Any more than this, and
# it'll just use the default reconnect time from config
my @RECONN_DELAYS = qw( 0 0 5 20 60 );

sub new
{
   my $class = shift;
   my ( $core, $config ) = @_;

   my $self = bless {
   }, $class;

   my $irc = Net::Async::IRC->new(
      on_message      => sub { shift; $self->on_irc_message( @_ ) },
      on_ping_timeout => sub { shift; $self->on_ping_timeout( @_ ) },
      on_closed       => sub { shift; $self->on_closed( @_ ) },

      on_pong_reply => sub {
         my ( $self, $lag ) = @_;
         $core->log( "DEBUG", "<Conn>", sprintf "Received PONG reply in %.2fs", $lag );
      },

      encoding => "UTF-8",
   );

   $core->get_loop->add( $irc );
   $self->{irc} = $irc;

   $self->{core} = $core;
   $self->{config} = $config;

   $self->{server_index} = 0;
   $self->{connect_errors} = 0;

   $self->reconfigure( $config );

   return $self;
}

sub reconfigure
{
   my $self = shift;
   my ( $config ) = @_;

   my @servers = $config->get_list(
      'server',
      { port => '@port', host => '@host' },
      default => { port => 6667 }
   );

   $self->{servers} = \@servers;

   my $irc = $self->{irc};
   $irc->{pingtime} = $config->get_string( '@pingtime', default => 60 );
   $irc->{pongtime} = $config->get_string( '@pongtime', default => 10 );

   $self->{reconn_delay} = $config->get_string( '@reconn_delay', default => 180 );
}

sub connect
{
   my $self = shift;

   my $server = $self->{servers}->[$self->{server_index}];

   my $host = $server->{host};
   my $port = $server->{port};

   my $core = $self->{core};

   my $config = $self->{config};

   if( $self->{is_connected} ) {
      $core->err( "No point reconnecting $self as it's already connected" );
      return;
   }

   $core->log( "CONN", "<Conn>", "Connecting to $host:$port" );

   my $irc = $self->{irc};
   $irc->login(
      host => $host,
      service => $port,

      nick     => $config->get_string( '@nick', default => "Cel" ),
      user     => $config->get_string( '@user', default => $ENV{USER} || "user" ),
      realname => $config->get_string( '@realname', default => "realname" ),
      pass     => $config->get_string( '@pass', default => undef ),

      on_login => sub {
         my ( $sock ) = @_;

         $self->{is_connected} = 1;
         $self->{connect_errors} = 0;
      },

      on_connect_error => sub {
         $core->err( "Connect failed" );
         $self->on_connect_error;
      },

      on_resolve_error => sub {
         $core->err( "Cannot resolve $host:$port - $_[0]" );
         # Name resolution error; probably best to cycle to the next server in
         # the list just in case
         $self->on_connect_error( cycle => 1 );
      }
   );
}

sub is_connected
{
   my $self = shift;
   return $self->{is_connected};
}

sub on_connect_error
{
   my $self = shift;
   my ( %args ) = @_;

   my $core = $self->{core};

   my $reconn_delay = $self->{reconn_delay};

   if( $self->{connect_errors} < scalar @RECONN_DELAYS ) {
      $reconn_delay = $RECONN_DELAYS[$self->{connect_errors}];
   }

   $self->{connect_errors}++;

   # If there have been more than 3 errors, there may be a problem with this
   # server. Cycle it anyway
   if( $args{cycle} or $self->{connect_errors} > 3 ) {
      $self->{server_index}++;
      $self->{server_index} %= scalar @{ $self->{servers} };
   }

   if( $reconn_delay ) {
      $core->log( "CONN", "<Conn>: ", "Need to reconnect to server [$self->{server_index}] after $reconn_delay seconds" );
      my $core = $self->{core};
      my $loop = $core->get_loop;
      $loop->enqueue_timer(
         delay => $reconn_delay,
         code  => sub { $self->connect },
      );
   }
   else {
      $self->connect;
   }
}

sub on_closed
{
   my $self = shift;

   my $core = $self->{core};

   $self->{is_connected} = 0;

   if( $self->{on_disconnected} ) {
     $self->{on_disconnected}->();
     delete $self->{on_disconnected};
   }
   else {
      $core->err( "Unexpected disconnect" );
      $self->on_connect_error;
   }
}

sub disconnect
{
   my $self = shift;
   my %args = @_;

   my $message = $args{message};

   $self->send_message( "QUIT", undef, $message );

   $self->{on_disconnected} = $args{on_disconnected};
}

sub shutdown
{
   my $self = shift;

   my $core = $self->{core};
   my $loop = $core->get_loop;

   $self->disconnect(
      on_disconnected => sub { $loop->remove( $self->{conn} ) },
   );
}

sub on_irc_message
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   return if $hints->{handled};

   my $core = $self->{core};

   my $method = "on_message_$command";

   # First look for internal handling
   return 1 if $self->can( $method ) and $self->$method( $message );

   return 1 if $core->run_plugins( [ $method, "on_message" ], $command, $message, $hints );

   my @args = $message->args;
   $core->log( "INFO", "<Message>", "$command @args" ) unless $hints->{synthesized};
   return 0;
}

sub on_ping_timeout
{
   my $self = shift;

   my $core = $self->{core};

   $core->log( "ERR", "<Error>", "PONG timer expired - reconnecting" );
   
   # TODO: Close the old connection
   $self->connect;
}

sub is_me
{
   my $self = shift;
   my ( $nick ) = @_;
   return $self->is_nick_me( $nick );
}

sub get_users_plugin
{
   my $self = shift;
   my $core = $self->{core};
   return $core->get_plugin( "users" );
}

sub get_user
{
   my $self = shift;
   my ( $nick ) = @_;
   return $self->get_users_plugin->get_user( $nick );
}

sub get_user_me
{
   my $self = shift;
   return $self->get_user( $self->{irc}->nick );
}

sub get_users
{
   my $self = shift;
   return $self->get_users_plugin->get_users;
}

sub nick_to_user
{
   my $self = shift;
   my ( $nick ) = @_;
   return $self->get_users_plugin->nick_to_user( $nick );
}

sub hints_to_user
{
   my $self = shift;
   return $self->get_users_plugin->hints_to_user( @_ );
}

sub get_channels_plugin
{
   my $self = shift;
   my $core = $self->{core};
   return $core->get_plugin( "channels" );
}

sub get_channel
{
   my $self = shift;
   my ( $name ) = @_;
   return $self->get_channels_plugin->get_channel( $name );
}

sub get_channels
{
   my $self = shift;
   return $self->get_channels_plugin->get_channels;
}

# There are a number of methods we want to reflect to the irc object
foreach my $method ( qw(
      send_message
      casefold_name
      is_nick_me
      cmp_prefix_flags
      cmp_prefix_modes
      prefix_mode2flag
      prefix_flag2mode
      isupport
      ) ) {
   no strict 'refs';
   *{$method} = sub { 
      my $self = shift;
      my $irc = $self->{irc};
      $irc->$method( @_ )
   };
}

###
# Commands
###

sub register_commands
{
   my $self = shift;
   my ( $commands_plugin ) = @_;

   $commands_plugin->register(
      plugin  => $self,
      command => "reconnect",
      perm    => 'master', # Allow non-owner masters

      args => [
         CelBot::Commands::ArgSpec->new( 'quitmsg', eatall => 1, optional => 1 ),
      ],

      summary => "Disconnect then reconnect to the IRC server",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "cycle",
      perm    => 'master', # Allow non-owner masters

      args => [
         CelBot::Commands::ArgSpec->new( 'quitmsg', eatall => 1, optional => 1 ),
      ],

      summary => "Disconnect then reconnect to the next IRC server in the list",
   );
}

sub command_reconnect
{
   my $self = shift;
   my ( $contect, $quitmsg ) = @_;

   $quitmsg = "Reconnecting..." if !defined $quitmsg;

   $self->disconnect(
      message => $quitmsg,
      on_disconnected => sub { $self->connect },
   );

   return ( "OK" );
}

sub command_cycle
{
   my $self = shift;
   my ( $contect, $quitmsg ) = @_;

   $self->{server_index}++;
   $self->{server_index} %= scalar @{ $self->{servers} };

   $quitmsg = "Reconnecting..." if !defined $quitmsg;

   $self->disconnect(
      message => $quitmsg,
      on_disconnected => sub { $self->connect },
   );

   return ( "OK" );
}

###
# Settings
###

sub register_settings
{
   my $self = shift;
   my ( $settings_plugin ) = @_;

   $settings_plugin->register(
      type => 'custom',
      name => "servers",
      perm => 'owner',

      storage => $self->{servers},

      print => sub {
         return join( ", ", map { $_->{host} . ":" . $_->{port} } @_ );
      },

      parse => sub {
         my ( $context ) = @_;

         my $text = $context->text;

         my ( $host, $port );
         if( $text =~ m/^(.*):(\d+)$/ ) {
            ( $host, $port ) = ( $1, $2 );
         }
         else {
            ( $host, $port ) = ( $text, 6667 );
         }

         return { host => $host, port => $port };
      },

      desc => "List of IRC servers to connect to",
   );
}

### Specific message handlers

sub on_message_001 # Initial connect
{
   my $self = shift;

   return 0; # Don't eat it
}

# Keep perl happy; keep Britain tidy
1;
