package CelBot::Plugin::AccountManagement;

use strict;
use constant PLUGIN_TYPE => "account-management";

use Carp;

sub new
{
   my $class = shift;
   my ( $core, $config ) = @_;

   my $self = bless {
      core => $core,
      irc  => $core->get_plugin( "irc" ),
      userdb => $core->get_plugin( "userdb" ),
   }, $class;

   return $self;
}

###
# Commands
###

sub register_commands
{
   my $self = shift;
   my ( $commands_plugin ) = @_;

   # First clear out all the existing ones
   $self->{commands} = \my %commands;
   $self->{privs}    = \my %privs;

   my $core = $self->{core};
   $core->run_plugins( "register_usermgmt_commands", $self );

   # Now we have privs - generate 'grant' and 'revoke' commands
   if( keys %privs ) {
      my %grantcmds;
      my %revokecmds;

      foreach my $priv ( keys %privs ) {
         my $grantcmd;
         my $revokecmd;

         my %args = %{ $privs{$priv} };

         if( $args{appliesto} eq "channels" ) {
            my $irc = $self->{irc};
            my $channels = $irc->get_channels_plugin;

            $args{args} = [
               CelBot::Commands::ArgSpec::Options->new(
                  {
                     in => $channels->gen_argspec,
                  }
               ),
            ];

            $grantcmd = { %args,
               function => sub {
                  my $context = shift;
                  $self->priv_delegate_channels( $context, $priv, 1, @_ );
               },
            };
            $revokecmd = { %args,
               function => sub {
                  my $context = shift;
                  $self->priv_delegate_channels( $context, $priv, undef, @_ );
               },
            };
         }
         else {
            $grantcmd  = \%args;
            $revokecmd = \%args;
         }

         $grantcmds{$priv}  = $grantcmd;
         $revokecmds{$priv} = $revokecmd;
      }

      $self->register_command(
        plugin   => $self,
        command  => "privs",
        perm     => 'recognised|self', # We'll do further checks later
        needhandle => 1,

        args => [],

        summary => "List privileges granted to a user",
      );

      $self->register_command(
         plugin  => $self,
         command => "grant",
         perm    => 'recognised', # We'll do further checks later
         needhandle => 1,

         funcname => "command_grant_priv",

         subcmds => \%grantcmds,

         summary => "Grant a privilege to a user",
      );

      $self->register_command(
         plugin  => $self,
         command => "revoke",
         perm    => 'recognised', # We'll do further checks later
         needhandle => 1,

         funcname => "command_revoke_priv",

         subcmds => \%revokecmds,

         summary => "Revoke a privilege from a user",
      );
   }

   # Now we have commands - work out what is valid for 'myacc' and 'useracc'
   my %mycmds;
   my %usercmds;

   foreach my $cmdname ( keys %commands ) {
      my $cmdstruct = $commands{$cmdname};

      my @perms = split( m{\|}, $cmdstruct->{perm} );

      if( grep { $_ eq "public" } @perms ) {
         $mycmds{$cmdname} = { %$cmdstruct, perm => "public" };
      }
      elsif( grep { $_ eq "self" } @perms ) {
         $mycmds{$cmdname} = { %$cmdstruct, perm => "recognised" };
      }

      $usercmds{$cmdname} = $cmdstruct;
   }

   $commands_plugin->register(
      plugin  => $self,
      command => "myacc",
      perm    => 'recognised',

      funcname => "command_useracc",

      subcmds => \%mycmds,
      default => "help",

      checkfunc => sub { $self->check_myacc( @_ ) },

      summary => "Manage your user account",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "useracc",
      perm    => 'recognised',

      subcmds => \%usercmds,
      default => "help",

      checkfunc => sub { $self->check_useracc( @_ ) },

      summary => "Manage other users' accounts",
   );
}

sub register_command
{
   my $self = shift;
   my %args = @_;

   my $command = delete $args{command} or croak "Need a 'command'";

   exists $self->{commands}->{$command} and croak "Usermanagement command '$command' already registered";

   $args{needhandle} = 1 unless defined $args{needhandle};

   $self->{commands}->{$command} = \%args;
}

sub register_priv
{
   my $self = shift;
   my %args = @_;

   my $priv = delete $args{priv} or croak "Need a 'priv'";

   $self->{privs}->{$priv} = \%args;
}

sub check_myacc
{
   my $self = shift;
   my ( $context, $cmdstruct, $substruct ) = @_;

   my $user = $context->cmduser;

   if( !defined $user ) {
      $context->respond_noise( "You cannot run this account management command" );
      return 0;
   }

   my $handle = $user->{handle};

   if( !defined $handle ) {
      $context->respond_noise( "You cannot run this account management command because you are not recognised" );
      return 0;
   }

   $context->{handle} = $handle;

   return 1;
}

sub check_useracc
{
   my $self = shift;
   my ( $context, $cmdstruct, $substruct ) = @_;

   my $needhandle = $substruct->{needhandle};

   return 1 unless $needhandle;

   my $handle = $context->pull_token;

   if( !defined $handle ) {
      $context->respond_noise( "Need to supply a handle after the command name" );
      return 0;
   }

   my $userdb = $self->{userdb};
   if( not $userdb->exists_handle( $handle ) ) {
      $context->respond_noise( "No such handle '$handle'" );
      return 0;
   }

   $context->{handle} = $handle;

   return 1;
}

sub priv_delegate_channels
{
   my $self = shift;
   my ( $context, $priv, $value, $options ) = @_;

   # This is a generic function for both grant and revoke; need some words for
   # messages
   my $verb_imp  = $value ? "grant" : "revoke";
   my $verb_past = $value ? "granted" : "revoked";

   my $handle = $context->{handle};

   my $core = $self->{core};

   my $ret;

   if( my $chan = $options->{in} ) {
      my $channame = $chan->name;
      my $already_has;

      if( !$value ) {
         $already_has = $core->ask_plugins( "ask_handle_global_data", $handle, $priv );
         if( $already_has ) {
            $context->respond( "Revoking $priv from $handle in $channame will not take effect as it is granted globally" );
         }
      }

      $already_has = $core->ask_plugins( "ask_handle_channel_data", $handle, $channame, $priv );
      if( $already_has and $value ) {
         return "$handle already has $priv in $channame";
      }
      elsif( !$already_has and !$value ) {
         return "$handle does not have $priv in $channame";
      }

      $core->run_plugins( "do_handle_set_channel_data", $handle, $channame, $priv, $value );
      $ret = ucfirst($verb_past) . " $handle $priv in $channame";
   }
   else {
      my $already_has = $core->ask_plugins( "ask_handle_global_data", $handle, $priv );
      if( $already_has and $value ) {
         return "$handle already has $priv globally";
      }
      elsif( !$already_has and !$value ) {
         return "$handle does not have $priv globally";
      }

      $core->run_plugins( "do_handle_set_global_data", $handle, $priv, $value );
      $ret = ucfirst($verb_past) . " $handle $priv globally";
   }

   my $privstruct = $self->{privs}->{$priv};
   if( my $func = $privstruct->{after_priv} ) {
      $func->( $context, $value, $options );
   }

   return $ret;
}

sub get_useracc_privs
{
   my $self = shift;
   my ( $handle ) = @_;

   my $core = $self->{core};

   my @global_privs;
   my %channel_privs;

   foreach my $priv ( sort keys %{ $self->{privs} } ) {
      my $chanhash = $core->ask_plugins( "ask_handle_channels_data", $handle, $priv );
      push @global_privs, $priv if delete $chanhash->{global};
      foreach my $channel ( keys %$chanhash ) {
         push @{ $channel_privs{$channel} }, $priv if $chanhash->{$channel};
      }
   }

   return \@global_privs, \%channel_privs;
}

sub command_useracc_privs
{
   my $self = shift;
   my ( $context ) = @_;

   my $handle = $context->{handle};

   my ( $global_privs, $channel_privs ) = $self->get_useracc_privs( $handle );

   my @ret = @$global_privs ? "Global privs for $handle: " . join( " ", @$global_privs )
                            : "No global privs for $handle";

   foreach my $chan ( sort keys %$channel_privs ) {
      push @ret, "  $chan: " . join( " ", @{ $channel_privs->{$chan} } ) if @{ $channel_privs->{$chan} };
   }

   return @ret;
}

sub get_useracc_for_display
{
   my $self = shift;
   my ( $handle, $ret ) = @_;

   my ( $global_privs, $channel_privs ) = $self->get_useracc_privs( $handle );

   my @ret = @$global_privs ? "Privs: " . join( " ", @$global_privs )
                            : "No global privs";

   foreach my $chan ( sort keys %$channel_privs ) {
      push @ret, "  $chan: " . join( " ", @{ $channel_privs->{$chan} } ) if @{ $channel_privs->{$chan} };
   }

   $ret->{privs} = \@ret;
   return 1;
}

###
# Internal Commands
###

sub register_usermgmt_commands
{
   my $self = shift;
   my ( $usermgmt_plugin ) = @_;

   $usermgmt_plugin->register_command(
      plugin  => $self,
      command => "help",
      perm    => 'public',
      needhandle => 0,

      args => [],

      summary => "Obtain additional help on user account management commands",
   );
}

sub command_useracc_help
{
   my $self = shift;

   my $commands = $self->{commands};

   return (
      q{Manage a user account},
      q{Usage: myacc $command $args...},
      q{       useracc $command [$handle] $args...},
      q{For more information, see 'help myacc' or 'help useracc'},
   );
}

# Keep perl happy, keep Britain tidy
1;
