package CelBot::Plugin::AutoMode;

use strict;
use constant PLUGIN_TYPE => "automode";

sub new
{
   my $class = shift;
   my ( $core, $config ) = @_;

   my $self = bless {
      core => $core,
      irc  => $core->get_plugin( "irc" ),
   }, $class;

   return $self;
}

my %mode_priv = (
   o => "op",
   h => "halfop",
   v => "voice",
);

sub channel_occupant
{
   my $self = shift;
   my ( $channel, $user ) = @_;

   my $irc = $self->{irc};

   my $me = $irc->get_user_me;
   my $myflag = $channel->get_userflag( $me );

   my $i_am_op = $irc->cmp_prefix_flags( $myflag, '@' ) >= 0;
   return unless $i_am_op;

   my $userflag = $channel->get_userflag( $user );

   my $core = $self->{core};

   foreach my $mode ( qw( o h v ) ) {
      next unless $core->ask_plugins( "permit_user_channel_priv", $user, $channel, $mode_priv{$mode} );

      my $flag = $irc->prefix_mode2flag( $mode );
      if( $irc->cmp_prefix_flags( $userflag, $flag ) < 0 ) {
         $channel->mode( "+$mode", $user->nick, 1 );
         last;
      }
   }
}

sub remode_channel
{
   my $self = shift;
   my ( $channel ) = @_;

   foreach my $occupant ( $channel->get_users ) {
      $self->channel_occupant( $channel, $occupant );
   }
}

sub remode_user
{
   my $self = shift;
   my ( $user ) = @_;

   foreach my $channel ( $user->channels ) {
      $self->channel_occupant( $channel, $user );
   }
}

sub on_channel_join
{
   my $self = shift;
   my ( $channel, $user ) = @_;

   $self->channel_occupant( $channel, $user );
}

sub on_channel_usermode
{
   my $self = shift;
   my ( $channel, $user, $sense, $flag ) = @_;

   return unless $user->is_me;

   my $irc = $self->{irc};

   my $myflag = $channel->get_userflag( $irc->get_user_me );
   my $i_am_op = $irc->cmp_prefix_flags( $myflag, '@' ) >= 0;

   return 1 unless $i_am_op;

   $self->remode_channel( $channel );

   return 1;
}

sub on_channel_who
{
   my $self = shift;
   my ( $channel, @who ) = @_;

   my $irc = $self->{irc};

   my $myflag = $channel->get_userflag( $irc->get_user_me );
   my $i_am_op = $irc->cmp_prefix_flags( $myflag, '@' ) >= 0;

   return 1 unless $i_am_op;

   $self->remode_channel( $channel );

   return 1;
}

sub on_recognise_user
{
   my $self = shift;
   my ( $user, $handle ) = @_;

   $self->remode_user( $user );
}

sub permit_user_channel_priv
{
   my $self = shift;
   my ( $user, $channel, $priv ) = @_;

   return 1 if $user->get_channel_data( $channel, $priv );
   return undef;
}

###
# Commands
###

sub register_commands
{
   my $self = shift;
   my ( $commands_plugin ) = @_;

   my $irc = $self->{irc};
   my $users    = $irc->get_users_plugin;
   my $channels = $irc->get_channels_plugin;

   $commands_plugin->register(
      plugin  => $self,
      command => "remodeuser",
      perm    => 'master',

      args => [
         $users->gen_argspec,
      ],

      summary => "Reapply automode for a given user",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "remodechan",
      perm    => 'master', # probably should be channel master one day when we have such ability

      args => [
         $channels->gen_argspec,
      ],

      summary => "Reapply automode for a given channel",
   );
}

sub command_remodeuser
{
   my $self = shift;
   my ( $context, $user ) = @_;

   $self->remode_user( $user );

   return ( "OK" );
}

sub command_remodechan
{
   my $self = shift;
   my ( $context, $channel ) = @_;

   $self->remode_channel( $channel );

   return ( "OK" );
}

sub register_usermgmt_commands
{
   my $self = shift;
   my ( $usermgmt_plugin ) = @_;

   my $core = $self->{core};

   my $after_priv = sub {
      my ( $context, $value, $options ) = @_;
      return unless $value;

      my $handle = $context->{handle};
      my @users = $core->ask_plugins_list( "ask_map_handle_to_users", $handle, 1 );

      return unless @users;

      if( my $channel = $options->{in} ) {
         $self->channel_occupant( $channel, $_ ) foreach @users;
      }
      else {
         $self->remode_user( $_ ) foreach @users;
      }
   };

   $usermgmt_plugin->register_priv(
      priv => "op",
      perm => 'master',

      appliesto => "channels",

      after_priv => $after_priv,

      summary => "Channel operator status",
   );

   $usermgmt_plugin->register_priv(
      priv => "halfop",
      perm => 'master',

      appliesto => "channels",

      after_priv => $after_priv,

      summary => "Channel half-operator status",
   );
}

# Keep perl happy, keep Britain tidy
1;
