package CelBot::Plugin::Hostmasks;

use strict;
use constant PLUGIN_TYPE => "hostmasks";

sub new
{
   my $class = shift;
   my ( $core, $config ) = @_;

   my $irc = $core->get_plugin( "irc" );

   my $self = bless {
      core => $core,
      irc  => $irc,
   }, $class;

   return $self;
}

sub scatter_user_verify_handle
{
   my $self = shift;
   my ( $gencb, $user, $handle ) = @_;

   my $core = $self->{core};
   my $hostmasks = $core->ask_plugins( "ask_handle_data", $handle, "hostmasks" );

   return 0 unless $hostmasks and @$hostmasks;

   my $location = $user->location;

   foreach my $hostmask ( @$hostmasks ) {
      my $re = mask_to_re( $hostmask );

      return 1 if $location =~ m/^$re$/;
   }

   return 0;
}

## Utility functions
sub mask_to_re
{
   my ( $mask ) = @_;

   my $re = $mask;
   $re =~ s{(\*\*|\*|\?|\.)}{
      $1 eq "**" and ".*"    or
      $1 eq "*"  and "[^.]*" or
      $1 eq "?"  and "."     or
      $1 eq "."  and "\\."
   }eg;

   return $re;
}

sub mask_to_re_for_mask
{
   my ( $mask ) = @_;

   my $re = $mask;
   $re =~ s{(\*\*|\*|\?|\.)}{
      $1 eq "**" and ".*"               or
      $1 eq "*"  and "[^.*]*\\*?[^.*]*" or
      $1 eq "?"  and "."                or
      $1 eq "."  and "\\."
   }eg;

   return $re;
}

sub on_handle_sync_from_user
{
   my $self = shift;
   my ( $handle, $user, $cmdcontext ) = @_;

   my $core = $self->{core};

   my $newmask = $user->location; # TODO: Some consideration on wildcards?

   my $hostmasks = $core->ask_plugins( "ask_handle_data", $handle, "hostmasks" ) || [];

   if( @$hostmasks ) {
      foreach my $hostmask ( @$hostmasks ) {
         my $re = mask_to_re_for_mask( $hostmask );
         if( $newmask =~ m/^$re$/ ) {
            undef $newmask;
            last;
         }
      }
   }

   if( defined $newmask ) {
      $cmdcontext->respond( "Adding hostmask '$newmask' to $handle" );
      push @$hostmasks, $newmask;
      $core->run_plugins( "do_handle_set_data", $handle, "hostmasks", $hostmasks );
   }

   return 1;
}

###
# Commands
###

sub register_usermgmt_commands
{
   my $self = shift;
   my ( $usermgmt_plugin ) = @_;

   $usermgmt_plugin->register_command(
      plugin  => $self,
      command => "hostmasks",
      perm    => 'self|master',

      subcmds => {
         get => { 
            args => [],
            summary => "Show current list of hostmasks",
         },

         add => { 
            args => [ CelBot::Commands::ArgSpec::Bareword->new( 'mask' ) ],
            summary => "Add a new hostmask to the list, merging if necessary",
         },

         del => {
            args => [ CelBot::Commands::ArgSpec::Bareword->new( 'mask' ) ],
            summary => "Remove existing hostmask(s) from the list",
         },

         delall => {
            args => [],
            summary => "Remove all existing hostmask(s) from the list",
         },
      },
      default => "get",

      summary => "Adjust recognition hostmasks",
   );
}

sub command_useracc_hostmasks_get
{
   my $self = shift;
   my ( $context ) = @_;

   my $handle = $context->{handle};

   my $core = $self->{core};
   my $hostmasks = $core->ask_plugins( "ask_handle_data", $handle, "hostmasks" );

   return "No hostmasks for $handle" unless defined $hostmasks;
   return "Hostmasks for $handle: " . join( ", ", @$hostmasks );
}

sub command_useracc_hostmasks_add
{
   my $self = shift;
   my ( $context, $newmask ) = @_;

   my $handle = $context->{handle};

   my $core = $self->{core};
   my $hostmasks = $core->ask_plugins( "ask_handle_data", $handle, "hostmasks" ) || [];

   if( @$hostmasks ) {
      foreach my $hostmask ( @$hostmasks ) {
         my $re = mask_to_re_for_mask( $hostmask );
         if( $newmask =~ m/^$re$/ ) {
            return "Not adding new mask $newmask as it is already matched by $hostmask";
         }
      }

      if( $newmask =~ m/[*?]/ ) {
         # Now see if we can remove any of the existing ones
         my $newre = mask_to_re_for_mask( $newmask );
         @$hostmasks = grep {
            if( $_ =~ m/^$newre$/ ) {
               $context->respond( "Removing mask $_ as it is covered by $newmask" );
               0
            }
            else {
               1
            }
         } @$hostmasks;
      }
   }

   push @$hostmasks, $newmask;

   $core->run_plugins( "do_handle_set_data", $handle, "hostmasks", $hostmasks );

   $core->log( "AUDIT", "<Hostmasks>", ($context->cmduser||"[console]") . " changed $handle hostmasks to: " . join( ", ", @$hostmasks ) );

   my @users = $core->ask_plugins_list( "ask_map_handle_to_users", $handle, 0 );
   foreach my $user ( @users ) {
      $user->{rerecognise} = 1;
      my $irc = $self->{irc};
      $irc->send_message( "WHOIS", undef, $user->nick );
   }

   return "New hostmasks for $handle: " . join( ", ", @$hostmasks );
}

sub command_useracc_hostmasks_del
{
   my $self = shift;
   my ( $context, $oldmask ) = @_;

   my $handle = $context->{handle};

   my $core = $self->{core};
   my $hostmasks = $core->ask_plugins( "ask_handle_data", $handle, "hostmasks" );

   if( !$hostmasks or !@$hostmasks ) {
      return "No hostmasks to delete";
   }

   # TODO: Consideration of masks?
   @$hostmasks = grep { $_ ne $oldmask } @$hostmasks;

   $core->run_plugins( "do_handle_set_data", $handle, "hostmasks", $hostmasks );

   $core->log( "AUDIT", "<Hostmasks>", ($context->cmduser||"[console]") . " changed $handle hostmasks to: " . join( ", ", @$hostmasks ) );

   return "New hostmasks for $handle: " . join( ", ", @$hostmasks );
}

sub command_useracc_hostmasks_delall
{
   my $self = shift;
   my ( $context ) = @_;

   my $handle = $context->{handle};

   my $core = $self->{core};
   my $hostmasks = $core->ask_plugins( "ask_handle_data", $handle, "hostmasks" );

   if( !$hostmasks or !@$hostmasks ) {
      return "No hostmasks to delete";
   }

   @$hostmasks = ();

   $core->run_plugins( "do_handle_set_data", $handle, "hostmasks", $hostmasks );

   $core->log( "AUDIT", "<Hostmasks>", ($context->cmduser||"[console]") . " changed $handle hostmasks to: " . join( ", ", @$hostmasks ) );

   return "New hostmasks for $handle: " . join( ", ", @$hostmasks );
}

sub get_useracc_for_display
{
   my $self = shift;
   my ( $handle, $ret ) = @_;

   my $core = $self->{core};
   my $hostmasks = $core->ask_plugins( "ask_handle_data", $handle, "hostmasks" );

   if( $hostmasks and @$hostmasks ) {
      $ret->{hostmasks} = "Hostmasks: " . join( ", ", @$hostmasks );
   }
   else {
      $ret->{hostmasks} = "No hostmasks";
   }

   return 1;
}

# Keep perl happy; keep Britain tidy
1;
