package CelBot::Plugin::Infoline;

use strict;
use constant PLUGIN_TYPE => "infoline";

sub new
{
   my $class = shift;
   my ( $core, $config ) = @_;

   my $self = bless {
      core => $core,
   }, $class;

   return $self;
}

sub print_user_infoline
{
   my $self = shift;
   my ( $user, $channel ) = @_;

   my $infoline = $user->get_channel_data( $channel, "info" );

   my $nick = $user->nick;

   if( defined $infoline ) {
      $channel->notice( "[$nick] $infoline" );
   }
}

sub on_channel_join
{
   my $self = shift;
   my ( $channel, $user ) = @_;

   $self->print_user_infoline( $user, $channel );
}

sub on_recognise_user
{
   my $self = shift;
   my ( $user, $handle ) = @_;

   my $now = time();

   foreach my $channel ( $user->channels ) {
      if( $now - $channel->get_jointime( $user ) < 10 ) {
         $self->print_user_infoline( $user, $channel );
      }
   }
}

###
# Commands
###

sub register_usermgmt_commands
{
   my $self = shift;
   my ( $usermgmt_plugin ) = @_;

   my $core = $self->{core};
   my $channels = $core->get_plugin( "channels" );

   $usermgmt_plugin->register_command(
      plugin  => $self,
      command => "infoline",
      perm    => 'self|master',

      subcmds => {
         get => {
            args => [],
            summary => "Show current infolines",
         },

         set => {
            args => [
               CelBot::Commands::ArgSpec::Bareword->new( 'text', eatall => 1 ),
            ],
            summary => "Set a global default infoline",
         },

         clear => {
            args => [],
            summary => "Remove global default infoline",
         },

         setchan => {
            args => [
               $channels->gen_argspec,
               CelBot::Commands::ArgSpec::Bareword->new( 'text', eatall => 1 ),
            ],
            summary => "Set a per-channel infoline",
         },

         clearchan => {
            args => [
               $channels->gen_argspec,
            ],
            summary => "Remove a per-channel infoline",
         },
      },
      default => "get",

      summary => "Adjust channel announcement infolines",
   );
}

sub command_useracc_infoline_get
{
   my $self = shift;
   my ( $context ) = @_;

   my $handle = $context->{handle};

   my $core = $self->{core};

   my $infolines = $core->ask_plugins( "ask_handle_channels_data", $handle, "info" );

   my @ret;

   my $global = delete $infolines->{global};
   push @ret, defined $global ? "Global: $global" : "No global";

   push @ret, "$_: $infolines->{$_}" for sort keys %$infolines;

   return @ret;
}

sub command_useracc_infoline_set
{
   my $self = shift;
   my ( $context, $text ) = @_;

   my $handle = $context->{handle};

   my $core = $self->{core};

   $core->run_plugins( "do_handle_set_global_data", $handle, "info", $text ) or
      return "Not able to set data";

   return "Set global infoline for $handle to '$text'";
}

sub command_useracc_infoline_clear
{
   my $self = shift;
   my ( $context ) = @_;

   my $handle = $context->{handle};

   my $core = $self->{core};

   $core->run_plugins( "do_handle_set_global_data", $handle, "info", undef ) or
      return "Not able to set data";

   return "Cleared global infoline for $handle";
}

sub command_useracc_infoline_setchan
{
   my $self = shift;
   my ( $context, $channel, $text ) = @_;

   my $handle = $context->{handle};
   my $channame = $channel->name;

   my $core = $self->{core};

   $core->run_plugins( "do_handle_set_channel_data", $handle, $channame, "info", $text ) or
      return "Not able to set data";

   return "Set channel infoline in $channame for $handle to '$text'";
}

sub command_useracc_infoline_clearchan
{
   my $self = shift;
   my ( $context, $channel ) = @_;

   my $handle = $context->{handle};
   my $channame = $channel->name;

   my $core = $self->{core};

   $core->run_plugins( "do_handle_set_channel_data", $handle, $channame, "info", undef ) or
      return "Not able to set data";

   return "Cleared channel infoline in $channame for $handle";
}

sub get_useracc_for_display
{
   my $self = shift;
   my ( $handle, $ret ) = @_;

   my $core = $self->{core};

   my $infolines = $core->ask_plugins( "ask_handle_channels_data", $handle, "info" );

   my @ret;

   my $global = delete $infolines->{global};
   push @ret, defined $global ? "Infoline: $global" : "No global";

   push @ret, "  $_: $infolines->{$_}" for sort keys %$infolines;

   $ret->{infoline} = \@ret;

   return 1;
}

# Keep perl happy, keep Britain tidy
1;
