package CelBot::Plugin::Recognise;

use strict;
use constant PLUGIN_TYPE => "recognise";

use CPS qw( kforeach );

sub new
{
   my $class = shift;
   my ( $core, $config ) = @_;

   my $irc = $core->get_plugin( "irc" );

   my $self = bless {
      core => $core,
      irc  => $irc,
   }, $class;

   return $self;
}

sub recognise
{
   my $self = shift;
   my ( $user, $handle ) = @_;

   # Early exit if no change
   return if !defined $handle and !defined $user->{handle};
   return if  defined $handle and  defined $user->{handle} and $handle eq $user->{handle};

   $user->{handle} = $handle;

   my $core = $self->{core};
   $core->run_plugins_after( "on_recognise_user", $user, $handle );
}

# FUNCTION
sub shorten_nick
{
   my ( $nick ) = @_;

   $nick =~ s/[\[\|\]].*// or return undef;
   return $nick;
}

sub ask_map_user_to_handles
{
   my $self = shift;
   my ( $user ) = @_;

   my $core = $self->{core};

   my $nick = $user->nick_folded;

   my @handles = $core->ask_plugins_list( "ask_map_nick_to_handles", $nick );

   if( !@handles and my $shortnick = shorten_nick( $nick ) ) {
      # Users often like to put a status message in their nick; usually
      # Foo|Away or Foo[Away]. Try to cope with those
      @handles = $core->ask_plugins_list( "ask_map_nick_to_handles", $shortnick );
   }

   return @handles;
}

sub ask_map_handle_to_users
{
   my $self = shift;
   my ( $handle, $only_recognised ) = @_;

   my $core = $self->{core};
   my $irc  = $self->{irc};

   if( $only_recognised ) {
      # Only want recognised ones - this is a big shortcut
      return grep { $_->{handle} eq $handle } $irc->get_users;
   }

   my $nicks = $core->ask_plugins( "ask_handle_data", $handle, "nicks" );
   return unless $nicks and @$nicks;

   my @users;
   foreach my $user ( $irc->get_users ) {
      my $nick = $user->nick_folded;
      my $shortnick;

      if( grep { $_ eq $nick } @$nicks ) {
         push @users, $user;
      }
      elsif( $shortnick = shorten_nick( $nick ) and grep { $_ eq $shortnick } @$nicks ) {
         push @users, $user;
      }
   }

   return @users;
}

sub try_recognise
{
   my $self = shift;
   my ( $user ) = @_;

   my $core = $self->{core};

   # Recognition is a two-stage process. First we map the IRC nick of the user
   # to a candidate list of possible DB handles. We then try to verify those
   # claims by whatever method is appropriate.
   my @handles = $core->ask_plugins_list( "ask_map_user_to_handles", $user );

   return unless @handles;

   $self->try_recognise_user_handles( $user, \@handles );
}

sub try_recognise_user_handles
{
   my $self = shift;
   my ( $user, $handles, $kdone ) = @_;

   my $core = $self->{core};

   my $recognised_handle;
   kforeach( $handles, sub {
      my ( $handle, $knext, $klast ) = @_;

      $core->scatter_plugins(
         sub {
            my %results = @_;

            # Find any plugins that said "yes"
            my @yes = grep { $results{$_} } keys %results;

            if( @yes ) {
               $core->log( "DEBUG", "<Recognise>", "Recognise $user as $handle because the following plugins said yes: " . join( ", ", @yes ) );
               $self->recognise( $user, $handle );
               $recognised_handle = $handle;
               $klast->();
            }
            else {
               $core->log( "DEBUG", "<Recognise>", "No plugin recognised $user as $handle" );
               $knext->();
            }
         }
         , "scatter_user_verify_handle", $user, $handle
      );
   }, sub {
      $kdone->( $recognised_handle ) if $kdone;
   } );
}

sub on_user_rename
{
   my $self = shift;
   my ( $user, $oldnick, $newnick ) = @_;

   $self->try_recognise( $user );
}

sub on_user_host
{
   my $self = shift;
   my ( $user, $ident, $host ) = @_;

   $self->try_recognise( $user );
}

sub on_handle_sync_from_user
{
   my $self = shift;
   my ( $handle, $user, $cmdcontext ) = @_;

   my $core = $self->{core};

   my $newnick = $user->nick_folded;

   my $nicks = $core->ask_plugins( "ask_handle_data", $handle, "nicks" ) || [];

   if( !grep { $_ eq $newnick } @$nicks ) {
      $cmdcontext->respond( "Adding nick '$newnick' to $handle" );
      push @$nicks, $newnick;
      $core->run_plugins( "do_handle_set_data", $handle, "nicks", $nicks );
   }

   # As a result of this addition, the user must now be recognised here
   $self->recognise( $user, $handle );

   return 1;
}

sub ask_user_data
{
   my $self = shift;
   my ( $user, $key ) = @_;

   my $handle = $user->{handle};
   return undef unless defined $handle;

   my $core = $self->{core};
   return $core->ask_plugins( "ask_handle_data", $handle, $key );
}

sub ask_user_global_data
{
   my $self = shift;
   my ( $user, $key ) = @_;

   my $handle = $user->{handle};
   return undef unless defined $handle;

   my $core = $self->{core};
   return $core->ask_plugins( "ask_handle_global_data", $handle, $key );
}

sub ask_user_channel_data
{
   my $self = shift;
   my ( $user, $channel, $key ) = @_;

   my $handle = $user->{handle};
   return undef unless defined $handle;

   my $core = $self->{core};
   return $core->ask_plugins( "ask_handle_channel_data", $handle, $channel->name, $key );
}

sub permit_command_perm
{
   my $self = shift;
   my ( $perm, $context ) = @_;

   my $user = $context->cmduser;
   return undef unless defined $user;

   foreach my $p ( split( m{\|}, $perm ) ) {
      # Any registered user may do 'recognised'
      return ( defined $user->{handle} ) if $p eq 'recognised';

      # Check predefined classes
      return ( $user->get_global_data( 'master' ) ) if $p eq 'master';
      return ( $user->get_global_data( 'owner' ) ) if $p eq 'owner';
   }

   return undef;
}

sub permit_command_nonpublic
{
   my $self = shift;
   my ( $context ) = @_;

   # Recognised users can perform nonpublic commands
   my $user = $context->cmduser;
   return undef unless defined $user;

   return 1 if defined $user->{handle};

   return undef;
}

sub permit_user_channel_priv
{
   my $self = shift;
   my ( $user, $channel, $priv ) = @_;

   # Autovoice all recognised users
   return 1 if $priv eq "voice" and defined $user->{handle};
   return undef;
}

###
# Commands
###

sub register_commands
{
   my $self = shift;
   my ( $commands_plugin ) = @_;

   my $irc = $self->{irc};
   my $users    = $irc->get_users_plugin;
   my $channels = $irc->get_channels_plugin;

   $commands_plugin->register(
      plugin  => $self,
      command => "whoami",
      perm    => 'public',

      args => [],

      summary => "Show the recognised user handle of the invoking user",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "iam",
      perm    => 'public',

      args => [
         CelBot::Commands::ArgSpec::Bareword->new( 'handle' ),
      ],

      summary => "Claim to be the given username and attempt to verify this claim",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "whois",

      args => [
         $users->gen_argspec,
      ],

      summary => "Show the recognised user handle of a given user",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "rerecognise",
      perm    => 'master',

      args => [
         CelBot::Commands::ArgSpec->new( 'users', eatall => 1 ),
      ],

      summary => "Forget recognition of users and discover again",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "recognise",
      perm    => 'owner',

      args => [
         $users->gen_argspec,
         CelBot::Commands::ArgSpec::Sugar->new( 'as' ),
         CelBot::Commands::ArgSpec::Bareword->new( 'handle' ),
      ],

      summary => "Force a user recognition",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "unrecognise",
      perm    => 'owner',

      args => [
         $users->gen_argspec,
      ],

      summary => "Remove a forced user recognition",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "regreport",
      perm    => 'master',

      args => [
         $channels->gen_argspec,
      ],

      summary => "Display a report about user registrations in a given channel",
   );
}

sub command_whoami
{
   my $self = shift;
   my ( $context ) = @_;

   my $cmduser = $context->cmduser;

   if( defined $cmduser->{handle} ) {
      return ( "You are $cmduser->{handle}" );
   }
   else {
      return ( "I do not recognise you" );
   }
}

sub command_iam
{
   my $self = shift;
   my ( $context, $handle ) = @_;

   my $cmduser = $context->cmduser;

   if( defined $cmduser->{handle} ) {
      return ( "You are already recognised as $cmduser->{handle}" );
   }
   else {
      $self->try_recognise_user_handles( $cmduser, [ $handle ], sub {
         my ( $recognised ) = @_;
         if( defined $recognised ) {
            $context->respond( "You are now recognised as $recognised" );
         }
         else {
            $context->respond( "I could not recognise you" );
         }
      } );
      return ();
   }
}

sub command_whois
{
   my $self = shift;
   my ( $context, $user ) = @_;

   my $ret = $user->nick . " (" . $user->location . ") is ";

   if( defined $user->{handle} ) {
      $ret .= "recognised as $user->{handle}";
   }
   else {
      $ret .= "not recognised";
   }

   ( $ret );
}

sub command_rerecognise
{
   my $self = shift;
   my ( $context, $users ) = @_;

   my $irc = $self->{irc};
   my $users_plugin = $irc->get_users_plugin;

   my @just_users = map { $irc->casefold_name( $_ ) } split( m/ /, $users );

   my $count = 0;

   # Flush all the user data caches
   foreach my $user ( $users_plugin->get_users ) {
      my $nick = $user->nick_folded;

      next if @just_users and !grep { $_ eq $nick } @just_users;

      $user->{rerecognise} = 1;

      # Now send a WHOIS message, who when the reply comes back we'll capture
      # the new ident and host. Though only if we were doing specific users
      $irc->send_message( "WHOIS", undef, $user->nick ) if @just_users;

      # TODO: We need some way to register "diversions" on numerics that come
      # back about this user, until such time as we find they're finished
      # Consider borrowing irssi-like behaviour

      $count++;
   }

   return ( "Forgot recognition of $count user" . ( $count == 1 ? "" : "s" ) );
}

sub command_recognise
{
   my $self = shift;
   my ( $context, $user, $handle ) = @_;

   defined $user->{handle} and return ( "User is already recognised as '$user->{handle}'" );

   $self->recognise( $user, $handle );

   return ( $user->nick . " is now recognised as $handle" );
}

sub command_unrecognise
{
   my $self = shift;
   my ( $context, $user ) = @_;

   defined $user->{handle} or return ( "User is not recognised" );

   $self->recognise( $user, undef );

   return ( $user->nick . " is no longer recognised" );
}

sub command_regreport
{
   my $self = shift;
   my ( $context, $channel ) = @_;

   my @users = $channel->get_users;

   my @verified;
   my @unverified;
   my @unrecognised;

   my $core = $self->{core};

   my $irc = $self->{irc};

   foreach my $user ( @users ) {
      next if $user->is_me;

      my @nicks;

      if( defined $user->{handle} ) {
         push @verified, $user->nick;
      }
      elsif( $core->ask_plugins_list( "ask_map_nick_to_handles", $user->nick_folded ) ) {
         push @unverified, $user->nick;
      }
      else {
         push @unrecognised, $user->nick;
      }
   }

   my $str = $channel->name . " has " .
      ( scalar @verified ) . " verified users, " .
      ( scalar @unverified ) . " unverified users, and " .
      ( scalar @unrecognised ) . " unrecognised users";

   return (
      $str,
      "Verified: " . join( " ", @verified ),
      "Unverified: " . join( " ", @unverified ),
      "Unrecognised: " . join( " ", @unrecognised ),
   );
}

sub register_usermgmt_commands
{
   my $self = shift;
   my ( $usermgmt_plugin ) = @_;

   $usermgmt_plugin->register_command(
      plugin  => $self,
      command => "nicks",
      perm    => 'self|master',

      subcmds => {
         get => {
            args => [],
            summary => "Show current list of nicks",
         },

         add => { 
            args => [ CelBot::Commands::ArgSpec::Bareword->new( 'nick' ) ],
            summary => "Add a new nick to the list",
         },

         del => {
            args => [ CelBot::Commands::ArgSpec::Bareword->new( 'nick' ) ],
            summary => "Remove existing nick from the list",
         },
      },
      default => "get",

      summary => "Adjust recognition nicks",
   );
}

sub command_useracc_nicks_get
{
   my $self = shift;
   my ( $context ) = @_;

   my $handle = $context->{handle};

   my $core = $self->{core};
   my $nicks = $core->ask_plugins( "ask_handle_data", $handle, "nicks" );

   return "No nicks for $handle" unless defined $nicks;
   return "Nicks for $handle: " . join( ", ", @$nicks );
}

sub command_useracc_nicks_add
{
   my $self = shift;
   my ( $context, $newnick ) = @_;

   $newnick = lc $newnick;

   my $handle = $context->{handle};

   my $core = $self->{core};
   my @clashes = $core->ask_plugins_list( "ask_map_nick_to_handles", $newnick );

   if( @clashes ) {
      return "Cannot add this nick as it clashes with other users";
   }

   my $nicks = $core->ask_plugins( "ask_handle_data", $handle, "nicks" ) || [];

   push @$nicks, $newnick;

   $core->run_plugins( "do_handle_set_data", $handle, "nicks", $nicks );

   $core->log( "AUDIT", "<Recognise>", ($context->cmduser||"[console]") . " changed $handle nicks to: " . join( ", ", @$nicks ) );

   my @users = $core->ask_plugins_list( "ask_map_handle_to_users", $handle, 0 );
   foreach my $user ( @users ) {
      $user->{rerecognise} = 1;
      my $irc = $self->{irc};
      $irc->send_message( "WHOIS", undef, $user->nick );
   }

   return "New nicks for $handle: " . join( ", ", @$nicks );
}

sub command_useracc_nicks_del
{
   my $self = shift;
   my ( $context, $oldnick ) = @_;

   my $handle = $context->{handle};

   my $core = $self->{core};
   my $nicks = $core->ask_plugins( "ask_handle_data", $handle, "nicks" );

   if( !$nicks or !@$nicks ) {
      return "No nicks to delete";
   }

   @$nicks = grep { $_ ne $oldnick } @$nicks;

   $core->run_plugins( "do_handle_set_data", $handle, "nicks", $nicks );

   $core->log( "AUDIT", "<Recognise>", ($context->cmduser||"[console]") . " changed $handle nicks to: " . join( ", ", @$nicks ) );

   return "New nicks for $handle: " . join( ", ", @$nicks );
}

sub get_useracc_for_display
{
   my $self = shift;
   my ( $handle, $ret ) = @_;

   my $core = $self->{core};
   my $nicks = $core->ask_plugins( "ask_handle_data", $handle, "nicks" );

   $ret->{recognise} = "Nicks: " . join( ", ", @$nicks );

   return 1;
}

# Keep perl happy; keep Britain tidy
1;
