package CelBot::Plugins;

use strict;

use base qw( Exporter );
our @EXPORT_OK = qw(
   construct_plugin
   reload_plugins
);

BEGIN {
   if( exists $INC{"constant.pm"} and not exists $INC{"Sub/Uplevel.pm"} ) {
      die "Loading Sub::Uplevel AFTER constant is too late for it to take effect\n";
   }

   use Sub::Uplevel;
}

use B qw( walksymtable svref_2object );
use Data::Compare;
use Module::PluginFinder 0.02;

my $pluginfinder = Module::PluginFinder->new(
   search_path => "CelBot::Plugin",
   typefunc    => "PLUGIN_TYPE",
);

sub get_subs_for_file
{
   my ( $file ) = @_;

   my @subs;

   # Avoid warning about 'Name B::GB::test_file_and_push used only once'
   no warnings 'once';
   local *B::GV::test_file_and_push = sub {
      my $gvref = shift;

      my $symname = $gvref->STASH->NAME . "::" . $gvref->NAME;

      my $cvref = $gvref->CV;
      defined $cvref or return;

      $cvref->isa( "B::CV" ) or return;

      # Don't find my own sub
      $symname eq "B::GV::test_file_and_push" and return;

      push @subs, $symname if $cvref->FILE eq $file;
   };

   walksymtable( \%::, 'test_file_and_push', sub { 1 }, "" );

   return @subs;
}

sub construct_plugin
{
   my ( $type, @args ) = @_;

   return $pluginfinder->construct( $type, @args );
}

our %last_mtime;

sub reload_plugins
{
   my $class = shift;
   $pluginfinder->rescan;

   foreach my $key ( keys %INC ) {
      next unless $key =~ m{^CelBot/};
      # The console doesn't like being reloaded
      next if $key =~ m{^CelBot/Console} or $key =~ m{^CelBot/Term};

      my $path = $INC{$key};

      my $mtime = (stat( $path ))[9];

      # Seen it already
      next if defined $last_mtime{$path} and $last_mtime{$path} == $mtime;

      # Not seen it but it's not changed since we started
      next if $mtime <= $^T;

      my @functions = get_subs_for_file( $path );

      no strict 'refs';

      my %oldf;
      foreach my $f ( @functions ) {
         my ( $p, $n ) = $f =~ m/^(.*::)(.*?)$/ or next;

         $oldf{$f} = *{$f}{CODE};
         # TODO: This deletes the entire glob, not just the CODE
         # Perhaps (if we wanted) we could preserve the other items?
         delete ${$p}{$n};
      }

      delete $INC{$key};

      eval {
         # Since get_subs_for_file() gets confused by 'use constant' coming
         # from the wrong file, we'll do something special when importing
         # constants; namely, don't reload them. They're constant.

         no warnings 'redefine';

         my $constant_import = \&constant::import;

         local *constant::import = sub {
            my $class = shift;
            return unless @_;

            my $callerpkg = caller;

            my %constants;
            if( $_[0] eq "HASH" ) {
               %constants = %{ $_[0] };
            }
            else {
               my $n = shift;
               $constants{$n} = @_ == 1 ? $_[0] : [ @_ ];
            }

            foreach my $n ( keys %constants ) {
               my $f = "${callerpkg}::$n";

               if( defined *{$f}{CODE} ) {
                  my $oldvalue = *{$f}{CODE}->();
                  unless( Compare( $oldvalue, $constants{$n} ) ) {
                     ::console_print_err( "Redefining $f to a different value\n" );
                     delete ${"${callerpkg}::"}{$n};
                  }
                  else {
                     delete $constants{$n};
                  }
               }
            }

            return unless keys %constants;
            
            uplevel 1, $constant_import, $class, \%constants;
         };

         require $key;
      };

      if( $@ ) {
         ::console_print_err( "Reloading $key failed - $@\n" );
         $INC{$key} = $path;

         # Put the old functions back
         foreach my $f ( keys %oldf ) {
            *{$f} = $oldf{$f};
         }
      }
      else {
         ::console_print_err( "Reloaded $key\n" );
         $last_mtime{$path} = $mtime;

         # %oldf will go out of scope anyway, so we don't need to do anything
      }
   }
}

# Keep perl happy; keep Britain tidy
1;
