package CelBot::Settings;

use strict;

use CelBot::Commands;

use Carp;

sub new
{
   my $class = shift;
   my ( $core ) = @_;

   my $self = bless {
      core => $core,
   }, $class;

   return $self;
}

sub register_commands
{
   my $self = shift;
   my ( $commands_plugin ) = @_;

   # First clear the existing settings
   $self->{settings} = {};

   my $helpargs = [
      CelBot::Settings::ArgSpec->new( 'setting', settings => $self->{settings} ),
   ];

   my $core = $self->{core};
   $core->run_plugins( "register_settings", $self );

   if( keys %{ $self->{settings} } ) {
      $commands_plugin->register(
         plugin  => $self,
         command => "get",
         perm    => 'master',

         args => [
            CelBot::Settings::ArgSpec->new( 'setting', settings => $self->{settings} ),
         ],

         summary => "Get the value of a setting",

         helpargs => $helpargs,
      );
   }

   if( grep { exists $_->{set} } values %{ $self->{settings} } ) {
      $commands_plugin->register(
         plugin  => $self,
         command => "set",
         perm    => 'master',

         args => [
            CelBot::Settings::ArgSpec->new( 'setting', settings => $self->{settings} ),
            CelBot::Commands::ArgSpec::Sugar->new( '=' ),
            CelBot::Commands::ArgSpec->new( 'value', trailing => 1 ),
         ],

         summary => "Set the value of a setting",

         helpargs => $helpargs,
      );
   }

   if( grep { exists $_->{add} } values %{ $self->{settings} } ) {
      $commands_plugin->register(
         plugin  => $self,
         command => "add",
         perm    => 'master',

         args => [
            CelBot::Settings::ArgSpec->new( 'setting', settings => $self->{settings} ),
            CelBot::Commands::ArgSpec::Sugar->new( '=' ),
            CelBot::Commands::ArgSpec->new( 'value', trailing => 1 ),
         ],

         summary => "Add a value to a list setting",

         helpargs => $helpargs,
      );
   }

   if( grep { exists $_->{ins} } values %{ $self->{settings} } ) {
      $commands_plugin->register(
         plugin  => $self,
         command => "ins",
         perm    => 'master',

         args => [
            CelBot::Settings::ArgSpec->new( 'setting', settings => $self->{settings} ),
            CelBot::Commands::ArgSpec::Bareword->new( 'index', match => qr/^\d+$/ ),
            CelBot::Commands::ArgSpec::Sugar->new( '=' ),
            CelBot::Commands::ArgSpec->new( 'value', trailing => 1 ),
         ],

         summary => "Insert a value into a list setting before the given index (0-based)",

         helpargs => $helpargs,
      );
   }

   if( grep { exists $_->{del} } values %{ $self->{settings} } ) {
      $commands_plugin->register(
         plugin  => $self,
         command => "del",
         perm    => 'master',

         args => [
            CelBot::Settings::ArgSpec->new( 'setting', settings => $self->{settings} ),
            CelBot::Commands::ArgSpec::Bareword->new( 'index', match => qr/^\d+$/ ),
         ],

         summary => "Delete the value from a list setting at the given index (0-based)",

         helpargs => $helpargs,
      );
   }
}

my %types = (
   custom  => {}, # callbacks given at 'register' time

   boolean => { print => \&print_boolean, parse => \&parse_boolean },
   'int'   => { print => \&print_scalar,  parse => \&parse_int },
   string  => { print => \&print_scalar,  parse => \&parse_string },
);

my %stores = (
   SCALAR => {
      get => sub {
         my ( $setting ) = @_;
         return ${ $setting->{storage} };
      },
      set => sub {
         my ( $setting, $value ) = @_;
         ${ $setting->{storage} } = $value;
      },
   },

   ARRAY => {
      get => sub {
         my ( $setting ) = @_;
         return @{ $setting->{storage} }
      },
      add => sub {
         my ( $setting, $value ) = @_;
         push @{ $setting->{storage} }, $value;
      },
      ins => sub {
         my ( $setting, $index, $value ) = @_;
         if( $index < scalar @{ $setting->{storage} } ) {
            splice @{ $setting->{storage} }, $index, 0, $value;
         }
         else {
            push @{ $setting->{storage} }, $value;
         }
      },
      del => sub {
         my ( $setting, $index ) = @_;
         die "No value at index $index\n" unless $index < @{ $setting->{storage} };
         splice @{ $setting->{storage} }, $index, 1, ();
      },
   },
);

sub register
{
   my $self = shift;
   my %args = @_;

   my $name = $args{name} or croak "Need a 'name'";

   my $type = delete $args{type} or croak "Need a 'type'";
   exists $types{$type} or croak "Unrecognised type '$type'";

   $args{perm} or croak "Need 'perm'";

   $self->{settings}->{$name} = \%args;

   my $storagetype = ref $args{storage};

   # Fetch parse/print and accessor methods, but let %args override
   %args = (
             %{ $types{$type} },
             exists $stores{$storagetype} ? %{ $stores{$storagetype} } : (),
             %args
           );

   $args{get} or croak "Need a 'get' method";
}

sub commandhelp_get
{
   my $self = shift;
   my ( $context, $setting ) = @_;

   my $name = $setting->{name};

   return "Setting '$name': $setting->{desc}";
}

sub command_get
{
   my $self = shift;
   my ( $context, $setting ) = @_;

   my $name = $setting->{name};

   unless( $context->allows_perm( $setting->{perm} ) ) {
      return ( "You are not allowed to get setting '$name'" );
   }

   eval {
      $setting->{before}->( 'get', $setting, $context ) if exists $setting->{before};
   };
   return "Cannot get $name: $@" if $@;

   return "$name: " . $setting->{print}->( $setting->{get}->( $setting ) );
}

sub commandhelp_set
{
   my $self = shift;
   my ( $context, $setting ) = @_;

   my $name = $setting->{name};

   if( !exists $setting->{set} ) {
      return "Cannot set $name";
   }

   return "Setting '$name': $setting->{desc}";
}

sub command_set
{
   my $self = shift;
   my ( $context, $setting ) = @_;

   my $name = $setting->{name};

   unless( $context->allows_perm( $setting->{perm} ) ) {
      return ( "You are not allowed to set setting '$name'" );
   }

   if( !exists $setting->{set} ) {
      return "Cannot set $name";
   }

   eval {
      $setting->{before}->( 'set', $setting, $context ) if exists $setting->{before};
      my $value = $setting->{parse}->( $context );
      $setting->{set}->( $setting, $value );
      $setting->{after}->( 'set', $setting, $context ) if exists $setting->{after};
   };
   return "Cannot set $name: $@" if $@;

   return "$name: " . $setting->{print}->( $setting->{get}->( $setting ) );
}

sub commandhelp_add
{
   my $self = shift;
   my ( $context, $setting ) = @_;

   my $name = $setting->{name};

   if( !exists $setting->{add} ) {
      return "Cannot add to $name";
   }

   return "Setting '$name': $setting->{desc}";
}

sub command_add
{
   my $self = shift;
   my ( $context, $setting ) = @_;

   my $name = $setting->{name};

   unless( $context->allows_perm( $setting->{perm} ) ) {
      return ( "You are not allowed to add to setting '$name'" );
   }

   if( !exists $setting->{add} ) {
      return "Cannot add to $name";
   }

   eval {
      $setting->{before}->( 'add', $setting, $context ) if exists $setting->{before};
      my $value = $setting->{parse}->( $context );
      $setting->{add}->( $setting, $value );
      $setting->{after}->( 'add', $setting, $context ) if exists $setting->{after};
   };
   return "Cannot add to $name: $@" if $@;

   return "$name: " . $setting->{print}->( $setting->{get}->( $setting ) );
}

sub commandhelp_ins
{
   my $self = shift;
   my ( $context, $setting ) = @_;

   my $name = $setting->{name};

   if( !exists $setting->{ins} ) {
      return "Cannot ins into $name";
   }

   return "Setting '$name': $setting->{desc}";
}

sub command_ins
{
   my $self = shift;
   my ( $context, $setting, $index ) = @_;

   my $name = $setting->{name};

   unless( $context->allows_perm( $setting->{perm} ) ) {
      return ( "You are not allowed to insert into setting '$name'" );
   }

   if( !exists $setting->{ins} ) {
      return "Cannot ins into $name";
   }

   eval {
      $setting->{before}->( 'ins', $setting, $context ) if exists $setting->{before};
      my $value = $setting->{parse}->( $context );
      $setting->{ins}->( $setting, $index, $value );
      $setting->{after}->( 'ins', $setting, $context ) if exists $setting->{after};
   };
   return "Cannot ins into $name: $@" if $@;

   return "$name: " . $setting->{print}->( $setting->{get}->( $setting ) );
}

sub commandhelp_del
{
   my $self = shift;
   my ( $context, $setting ) = @_;

   my $name = $setting->{name};

   if( !exists $setting->{del} ) {
      return "Cannot del from $name";
   }

   return "Setting '$name': $setting->{desc}";
}

sub command_del
{
   my $self = shift;
   my ( $context, $setting, $index ) = @_;

   my $name = $setting->{name};

   unless( $context->allows_perm( $setting->{perm} ) ) {
      return ( "You are not allowed to delete from setting '$name'" );
   }

   if( !exists $setting->{del} ) {
      return "Cannot del from $name";
   }

   eval {
      $setting->{before}->( 'del', $setting, $context ) if exists $setting->{before};
      $setting->{del}->( $setting, $index );
      $setting->{after}->( 'del', $setting, $context ) if exists $setting->{after};
   };
   return "Cannot del from $name: $@" if $@;

   return "$name: " . $setting->{print}->( $setting->{get}->( $setting ) );
}

### Handler methods for various types

sub print_scalar { $_[0] }

sub print_boolean { $_[0] ? "TRUE" : "FALSE" }

sub parse_boolean
{
   my ( $context ) = @_;

   my $value = $context->pull_token;
   defined $value or die "No value given for boolean setting\n";

   return 1 if $value eq "1" or lc $value eq "true";
   return 0 if $value eq "0" or lc $value eq "false";
   die "Unrecognised value '$value' for boolean setting\n";
}

sub parse_int
{
   my ( $context ) = @_;

   my $value = $context->pull_token;
   defined $value or die "No value given for integer setting\n";

   no warnings 'numeric';
   die "Unrecognised value '$value' for integer setting\n" unless $value eq int $value;

   return $value;
}

sub parse_string
{
   my ( $context ) = @_;

   my $value = $context->text;
   defined $value or $value = "";

   return $value;
}


# Keep perl happy, keep Britain tidy
1;


package CelBot::Settings::ArgSpec;

use strict;
use base qw( CelBot::Commands::ArgSpec::Bareword );

sub pull
{
   my $self = shift;
   my ( $context ) = @_;

   my $settingstruct = $self->SUPER::pull( $context );

   if( my $opts = $settingstruct->{options} ) {
      # Clone the settingstruct since we'll be modifying it
      $settingstruct = { %$settingstruct };

      while(1) {
         my $optname = $context->peek_token;
         last unless defined $optname;

         last unless exists $opts->{$optname};
         die "Cannot supply '$optname' option more than once\n" if exists $settingstruct->{$optname};

         my $opt = $opts->{$optname};
         die "You are not allowed to use the '$optname' option\n" if defined $opt->{perm} and not $context->allows_perm( $opt->{perm} );

         $context->pull_token; # Eat the token we peeked at

         $settingstruct->{$optname} = $opt->pull( $context );
      }
   }

   return $settingstruct;
}

sub validate
{
   my $self = shift;
   my ( $setting ) = @_;

   die "'$setting' is not a valid setting\n" unless exists $self->{settings}->{$setting};
   
   return $self->{settings}->{$setting};
}

1;
