package CelBot::User;

use strict;

use Carp;

use CelBot::Resolver;

use Socket qw( inet_aton inet_ntoa pack_sockaddr_in unpack_sockaddr_in AF_INET NI_NAMEREQD );

sub new
{
   my $class = shift;
   my ( $core, $nick ) = @_;

   my $irc = $core->get_plugin( "irc" );

   my $self = bless {
      core => $core,
      irc  => $irc,

      nick => $nick,
      nick_folded => $irc->casefold_name( $nick ),

      ident => undef,
      host_raw => undef, # Host as presented by IRCd, may be an IP address
      host  => undef,    # Host as resolved by us; may be equal to {host_raw} 
                         # or may have been resolved

      is_me => $irc->is_me( $nick ),

      module_data => {},
   }, $class;

   $core->run_plugins( "on_user_create", $self ) unless $self->{is_me};

   return $self;
}

use overload '""' => "STRING";

sub STRING
{
   my $self = shift;
   return __PACKAGE__."[nick=$self->{nick}]";
}

sub quit
{
   my $self = shift;
   my ( $reason ) = @_;

   my $core = $self->{core};
   $core->run_plugins( "on_user_quit", $self, $reason );
}

sub nick
{
   my $self = shift;
   return $self->{nick};
}

sub nick_folded
{
   my $self = shift;
   return $self->{nick_folded};
}

sub is_me
{
   my $self = shift;
   my $irc = $self->{irc};
   return $irc->is_me( $self->nick_folded );
}

sub ident
{
   my $self = shift;
   # Trim a leading '~' if there is one
   ( my $ident = $self->{ident} ) =~ s/^~//;
   return $ident;
}

sub host
{
   my $self = shift;
   return defined $self->{host} ? $self->{host} : $self->{host_raw};
}

sub location
{
   my $self = shift;

   my $ident = $self->ident;
   my $host  = $self->host;

   return "$ident\@$host";
}

sub change_nick
{
   my $self = shift;
   my ( $newnick ) = @_;

   my $core = $self->{core};
   my $irc = $self->{irc};

   my $oldnick = $self->{nick};

   return if $newnick eq $oldnick;

   $self->{nick} = $newnick;
   $self->{nick_folded} = $irc->casefold_name( $newnick );

   $core->run_plugins( "on_user_rename", $self, $oldnick, $newnick );
}

sub advise_host
{
   my $self = shift;
   my ( $ident, $host ) = @_;

   # Do we already know this?
   return if( (!$self->{rerecognise}) and 
              defined $self->{ident} and ( $self->{ident} eq $ident ) and
              defined $self->{host}  and ( $self->{host} eq $host or $self->{host_raw} eq $host ) );

   $self->{ident} = $ident;
   $self->{host_raw} = $host;

   delete $self->{rerecognise};

   my $core = $self->{core};

   unless( $host =~ m/^\d+\.\d+\.\d+\.\d+$/ ) {
      $self->_set_host( $host );
      return;
   }

   my $ip = $host;

   my $saddr = pack_sockaddr_in( 0, inet_aton( $ip ) );

   my $loop = $core->get_loop;

   CelBot::Resolver::resolve(
      loop => $loop,

      type => "getnameinfo",
      data => [ $saddr, NI_NAMEREQD ],

      on_resolved => sub {
         my ( $host, $service ) = @{ $_[0] };

         if( !defined $host ) {
            $self->_set_host( $ip );
            return;
         }

         CelBot::Resolver::resolve(
            loop => $loop,

            type => "getaddrinfo_hash",
            data => [ host => $host, family => AF_INET ],

            on_resolved => sub {
               my @res = @_;

               foreach my $res ( @res ) {
                  my $addr = $res->{addr};
                  my ( $saddr_port, $saddr_host ) = unpack_sockaddr_in( $addr );

                  my $saddr_ip = inet_ntoa( $saddr_host );

                  if( $saddr_ip eq $ip ) {
                     $self->_set_host( $host );
                     return;
                  }
               }

               # No candidates matched - resort to IP
               $self->_set_host( $ip );
            },

            on_error => sub {
               $core->err( "Forward resolution error on $host - $_[0]" );
               $self->_set_host( $ip );
            },
         );
      },

      on_error => sub {
         $core->err( "Resolution error on $ip - $_[0]" );
         $self->_set_host( $ip );
      },
   );
}

sub _set_host
{
   my $self = shift;
   my ( $host ) = @_;

   $self->{host} = $host;

   my $core = $self->{core};

   $core->run_plugins( "on_user_host", $self, $self->{ident}, $host );
}

sub store_module_data
{
   my $self = shift;
   my ( $module, $data ) = @_;

   $self->{module_data}->{$module} = $data;
}

sub retrieve_module_data
{
   my $self = shift;
   my ( $module ) = @_;

   return $self->{module_data}->{$module};
}

sub get_global_data
{
   my $self = shift;
   my ( $key ) = @_;

   my $core = $self->{core};
   return $core->ask_plugins( "ask_user_global_data", $self, $key );
}

sub get_channel_data
{
   my $self = shift;
   my ( $channel, $key ) = @_;

   my $core = $self->{core};
   return $core->ask_plugins( "ask_user_channel_data", $self, $channel, $key );
}

sub channels
{
   my $self = shift;

   # Don't try to store this in the object, as keeping it uptodate and
   # synchronised with the Channel objects is tricky. Instead, on the rare
   # occasions when it's needed, look it up dynamically.

   my $irc = $self->{irc};

   my @channels;
   foreach my $channel ( $irc->get_channels ) {
       push @channels, $channel if defined $channel->get_userflag( $self );
   }

   return @channels;
}

sub privmsg
{
   my $self = shift;
   my ( $text ) = @_;

   my $irc = $self->{irc};
   $irc->send_message( "PRIVMSG", undef, $self->{nick}, $text );
}

sub whois
{
   my $self = shift;
   my %args = @_;

   croak "Cannot whois - already in progress" if $self->{collect_whois};

   $self->{collect_whois} = {};
   $self->{on_whois} = $args{on_whois};

   my $irc = $self->{irc};
   $irc->send_message( "WHOIS", undef, $self->{nick} );
}

sub on_whois_message
{
   my $self = shift;
   my ( $message, $hints ) = @_;

   if( $self->{collect_whois} ) {
      $self->{collect_whois}->{$message->command} = [ $message->args ];

      if( $message->command eq "318" ) {
         $self->{on_whois}->( %{ $self->{collect_whois} } );

         delete $self->{collect_whois};
         delete $self->{on_whois};
      }
   }
}

# Keep perl happy, keep Britain tidy
1;
