package CelBot::Users;

use strict;

use CelBot::User;
use CelBot::Commands;

sub new
{
   my $class = shift;
   my ( $core ) = @_;

   my $self = bless {
      core => $core,
      irc  => $core->get_plugin( "irc" ),

      users => {},
   }, $class;

   return $self;
}

sub get_user
{
   my $self = shift;
   my ( $nick ) = @_;

   my $irc = $self->{irc};
   my $nick_folded = $irc->casefold_name( $nick );

   return $self->{users}->{$nick_folded};
}

sub get_users
{
   my $self = shift;
   return values %{ $self->{users} };
}

sub nick_to_user
{
   my $self = shift;
   my ( $nick ) = @_;

   my $irc = $self->{irc};
   my $nick_folded = $irc->casefold_name( $nick );

   return $self->{users}->{$nick_folded} ||= CelBot::User->new( $self->{core}, $nick );
}

sub hints_to_user
{
   my $self = shift;
   my ( $hints ) = @_;

   my $user = $self->nick_to_user( $hints->{prefix_nick} );

   $user->advise_host( $hints->{prefix_user}, $hints->{prefix_host} );

   return $user;
}

sub gen_argspec
{
   my $self = shift;

   if( @_ ) {
      my $name = shift;
      return CelBot::Users::ArgSpec->new( $name, users => $self, @_ );
   }
   else {
      # Cache a single one
      return $self->{argspec} ||= CelBot::Users::ArgSpec->new( 'user', users => $self );
   }
}

sub on_message_NICK
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   if( $hints->{prefix_is_me} ) {
      return 0;
   }

   my $user = $self->hints_to_user( $hints );

   my $oldnick_folded = $user->nick_folded;
   my $newnick = $message->arg(0);

   my $irc = $self->{irc};
   my $newnick_folded = $irc->casefold_name( $newnick );

   $self->{users}->{$newnick_folded} = delete $self->{users}->{$oldnick_folded};

   $user->change_nick( $newnick );

   return 1;
}

sub on_message_PRIVMSG
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   my $core = $self->{core};
   my $irc = $self->{irc};
   my $target = $message->arg(0);

   if( $irc->is_me( $target ) ) {
      my $user = $self->hints_to_user( $hints );
      $core->run_plugins( "on_user_privmsg", $user, $message->arg(1) );
      return 1;
   }

   return 0;
}

sub on_message_QUIT
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   if( $hints->{prefix_is_me} ) {
      return 0;
   }

   my $user = $self->hints_to_user( $hints );

   my $reason = $message->arg(0);
   defined $reason or $reason = "";

   $user->quit( $reason );

   delete $self->{users}->{$user->nick_folded};

   return 1;
}

sub on_message_311
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   my $user = $self->nick_to_user( $message->arg(1) );

   my $ident = $message->arg(2);
   my $host  = $message->arg(3);

   $user->advise_host( $ident, $host );

   $user->on_whois_message( $message, $hints );

   return 1;
}

# TODO: This list probably needs extending
my %WHOIS = map { $_ => 1 } qw( 307 311 312 313 317 318 319 );

sub on_message
{
   my $self = shift;
   my ( $command, $message, $hints ) = @_;

   if( $WHOIS{$command} ) {
      my $user = $self->nick_to_user( $message->arg(1) );
      $user->on_whois_message( $message, $hints );
   }

   return 0;
}

# If we're in channels, need some integration there

sub on_channel_self_join
{
   my $self = shift;
   my ( $channel ) = @_;

   # Need the ident/hostname of every user in the channel - request a WHO list
   $channel->start_who;

   return 1;
}

###
# Commands
###

sub register_commands
{
   my $self = shift;
   my ( $commands_plugin ) = @_;

   $commands_plugin->register(
      plugin  => $self,
      command => "as",
      perm    => 'owner',

      args => [
         $self->gen_argspec,
         CelBot::Commands::ArgSpec->new( 'command', trailing => 1 ),
      ],

      summary => "Execute a command as another user",
   );

   $commands_plugin->register(
      plugin  => $self,
      command => "privately",

      args => [
         CelBot::Commands::ArgSpec->new( 'command', trailing => 1 ),
      ],

      summary => "Redirect output of a command to a PRIVMSG to the invoking user",
   );
}

sub command_as
{
   my $self = shift;
   my ( $context, $asuser ) = @_;

   my $subcontext = $context->new_sub(
      cmduser => $asuser,

      cascade => 0,
   );

   my $core = $self->{core};

   my $cmds_plugin = $core->get_plugin( "commands" );
   $cmds_plugin->run_command( $subcontext );

   return ();
}

sub command_privately
{
   my $self = shift;
   my ( $context ) = @_;

   my $cmduser = $context->cmduser;
   unless( $cmduser ) {
      $context->respond_noise( "Cannot run a command privately to no user" );
      return ();
   }

   my $subcontext = $context->new_sub(
      scope   => "privmsg",

      max_spam => 10, # TODO: Per-user lookup?
      allows_noise => 1,

      responder => sub {
         $cmduser->privmsg( $_[0] );
      },
   );

   my $core = $self->{core};

   my $cmds_plugin = $core->get_plugin( "commands" );
   $cmds_plugin->run_command( $subcontext );

   return ();
}

# Keep perl happy, keep Britain tidy
1;


package CelBot::Users::ArgSpec;

use strict;
use base qw( CelBot::Commands::ArgSpec::Bareword );

sub validate
{
   my $self = shift;
   my ( $username ) = @_;

   my $user = $self->{users}->get_user( $username );

   die "No such user '$username'\n" if !defined $user;

   return $user;
}

1;
